from __future__ import annotations

from typing import cast

from universal_transfer_operator.constants import FileType, TransferMode
from universal_transfer_operator.data_providers import create_dataprovider
from universal_transfer_operator.data_providers.filesystem.base import BaseFilesystemProviders
from universal_transfer_operator.datasets.file.base import File
from universal_transfer_operator.universal_transfer_operator import TransferIntegrationOptions


def resolve_file_path_pattern(
    file: File,
    filetype: FileType | None = None,
    normalize_config: dict | None = None,
    transfer_params: TransferIntegrationOptions | None = None,
    transfer_mode: TransferMode = TransferMode.NONNATIVE,
) -> list[File]:
    """get file objects by resolving path_pattern from local/object stores
    path_pattern can be
    1. local location - glob pattern
    2. s3/gcs location - prefix

    :param file: File dataset object
    :param filetype: constant to provide an explicit file type
    :param normalize_config: parameters in dict format of pandas json_normalize() function
    :param transfer_params: kwargs to be used by method involved in transfer flow.
    :param transfer_mode: Use transfer_mode TransferMode; native, non-native or thirdparty.
    """
    if transfer_params is None:
        transfer_params = TransferIntegrationOptions()

    location = cast(
        BaseFilesystemProviders,
        create_dataprovider(
            dataset=file,
            transfer_params=transfer_params,
            transfer_mode=transfer_mode,
        ),
    )
    files = []
    for path in location.paths:
        if not path.endswith("/"):
            file = File(
                path=path,
                conn_id=file.conn_id,
                filetype=filetype,
                normalize_config=normalize_config,
            )
            files.append(file)
    if len(files) == 0:
        raise FileNotFoundError(f"File(s) not found for path/pattern '{file.path}'")
    return files



from __future__ import annotations

from typing import Any

from airflow.models import BaseOperator
from airflow.utils.context import Context

from universal_transfer_operator.constants import TransferMode
from universal_transfer_operator.data_providers import create_dataprovider, get_dataprovider_options_class
from universal_transfer_operator.datasets.file.base import File
from universal_transfer_operator.datasets.table import Table
from universal_transfer_operator.integrations import (
    get_conn_id,
    get_ingestor_option_class,
    get_transfer_integration,
)
from universal_transfer_operator.integrations.base import TransferIntegrationOptions


class UniversalTransferOperator(BaseOperator):
    """
    Transfers all the data that could be read from the source Dataset into the destination Dataset. From a DAG author
    standpoint, all transfers would be performed through the invocation of only the Universal Transfer Operator.

    :param source_dataset: Source dataset to be transferred.
    :param destination_dataset: Destination dataset to be transferred to.
    :param transfer_params: kwargs to be used by method involved in transfer flow.
    :param transfer_mode: Use transfer_mode TransferMode; native, non-native or thirdparty.
    :param if_exists: Overwrite file if exists. Default False.

    :return: returns the destination dataset
    """

    def __init__(
        self,
        *,
        source_dataset: Table | File,
        destination_dataset: Table | File,
        transfer_params: TransferIntegrationOptions | dict | None = None,
        transfer_mode: TransferMode = TransferMode.NONNATIVE,
        **kwargs,
    ) -> None:
        transfer_params = transfer_params or {}
        self.source_dataset = source_dataset
        self.destination_dataset = destination_dataset
        self.transfer_mode = transfer_mode
        self._transfer_params = transfer_params
        super().__init__(**kwargs)

    def execute(self, context: Context) -> Any:  # skipcq: PYL-W0613
        self._populate_transfer_params()

        if self.transfer_mode == TransferMode.THIRDPARTY:
            transfer_integration = get_transfer_integration(self.transfer_params)
            return transfer_integration.transfer_job(self.source_dataset, self.destination_dataset)

        source_dataprovider = create_dataprovider(
            dataset=self.source_dataset,
            transfer_params=self.transfer_params,
            transfer_mode=self.transfer_mode,
        )

        destination_dataprovider = create_dataprovider(
            dataset=self.destination_dataset,
            transfer_params=self.transfer_params,
            transfer_mode=self.transfer_mode,
        )
        native_path_available = destination_dataprovider.is_native_path_available(  # type: ignore
            source_dataset=self.source_dataset
        )
        if self.transfer_mode == TransferMode.NATIVE and native_path_available is False:
            raise ValueError(
                f"No native path available for {source_dataprovider.dataset} to"
                f" {destination_dataprovider.dataset}. You can try TransferMode - {TransferMode.NONNATIVE}"
                f" or {TransferMode.THIRDPARTY}"
            )

        destination_references = []
        for source_data in source_dataprovider.read():
            destination_references.append(destination_dataprovider.write(source_data))
        return destination_references

    def _populate_transfer_params(self):
        """
        Populate option class in transfer_params
        """
        self.transfer_params: TransferIntegrationOptions = (
            self._get_options_class(dataset=self.destination_dataset, transfer_params=self._transfer_params)(
                **self._transfer_params
            )
            if isinstance(self._transfer_params, dict)
            else self._transfer_params
        )

    def _get_options_class(
        self, transfer_params: TransferIntegrationOptions | dict, dataset: Table | File
    ) -> type[TransferIntegrationOptions]:
        """
        Get option class based on transfer type
        """
        if self.transfer_mode == TransferMode.THIRDPARTY:
            conn_id = get_conn_id(transfer_params=transfer_params)
            return get_ingestor_option_class(conn_id=conn_id)
        else:
            return get_dataprovider_options_class(dataset=dataset)



from __future__ import annotations

import socket

from airflow.providers.sqlite.hooks.sqlite import SqliteHook
from sqlalchemy import MetaData as SqlaMetaData, create_engine
from sqlalchemy.engine.base import Engine
from sqlalchemy.sql.schema import Table as SqlaTable

from universal_transfer_operator.data_providers.database.base import DatabaseDataProvider
from universal_transfer_operator.datasets.table import Metadata, Table
from universal_transfer_operator.universal_transfer_operator import TransferIntegrationOptions


class SqliteDataProvider(DatabaseDataProvider):
    """SqliteDataProvider represent all the DataProviders interactions with Sqlite Databases."""

    def __init__(
        self,
        dataset: Table,
        transfer_mode,
        transfer_params: TransferIntegrationOptions = TransferIntegrationOptions(),
    ):
        self.dataset = dataset
        self.transfer_params = transfer_params
        self.transfer_mode = transfer_mode
        self.transfer_mapping = set()
        self.LOAD_DATA_NATIVELY_FROM_SOURCE: dict = {}
        super().__init__(
            dataset=self.dataset, transfer_mode=self.transfer_mode, transfer_params=self.transfer_params
        )

    def __repr__(self):
        return f'{self.__class__.__name__}(conn_id="{self.dataset.conn_id})'

    @property
    def sql_type(self) -> str:
        return "sqlite"

    @property
    def hook(self) -> SqliteHook:
        """Retrieve Airflow hook to interface with the Sqlite database."""
        return SqliteHook(sqlite_conn_id=self.dataset.conn_id)

    @property
    def sqlalchemy_engine(self) -> Engine:
        """Return SQAlchemy engine."""
        # Airflow uses sqlite3 library and not SqlAlchemy for SqliteHook
        # and it only uses the hostname directly.
        airflow_conn = self.hook.get_connection(self.dataset.conn_id)
        return create_engine(f"sqlite:///{airflow_conn.host}")

    @property
    def default_metadata(self) -> Metadata:
        """Since Sqlite does not use Metadata, we return an empty Metadata instances."""
        return Metadata()

    # ---------------------------------------------------------
    # Table metadata
    # ---------------------------------------------------------
    @staticmethod
    def get_table_qualified_name(table: Table) -> str:
        """
        Return the table qualified name.

        :param table: The table we want to retrieve the qualified name for.
        """
        return str(table.name)

    def populate_metadata(self):  # skipcq: PTC-W0049
        """
        Since SQLite does not have a concept of databases or schemas, we just return the table as is,
        without any modifications.
        """
        pass

    def create_schema_if_needed(self, schema: str | None) -> None:
        """
        Since SQLite does not have schemas, we do not need to set a schema here.
        """

    def schema_exists(self, schema: str) -> bool:  # skipcq PYL-W0613,PYL-R0201
        """
        Check if a schema exists. We return false for sqlite since sqlite does not have schemas
        """
        return False

    def get_sqla_table(self, table: Table) -> SqlaTable:
        """
        Return SQLAlchemy table instance

        :param table: Astro Table to be converted to SQLAlchemy table instance
        """
        return SqlaTable(table.name, SqlaMetaData(), autoload_with=self.sqlalchemy_engine)

    @property
    def openlineage_dataset_name(self) -> str:
        """
        Returns the open lineage dataset name as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        Example: /tmp/local.db.table_name
        """
        conn = self.hook.get_connection(self.dataset.conn_id)
        return f"{conn.host}.{self.dataset.name}"

    @property
    def openlineage_dataset_namespace(self) -> str:
        """
        Returns the open lineage dataset namespace as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        Example: file://127.0.0.1:22
        """
        conn = self.hook.get_connection(self.dataset.conn_id)
        port = conn.port or 22
        return f"file://{socket.gethostbyname(socket.gethostname())}:{port}"

    @property
    def openlineage_dataset_uri(self) -> str:
        """
        Returns the open lineage dataset uri as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        return f"{self.openlineage_dataset_namespace}{self.openlineage_dataset_name}"



from __future__ import annotations

import glob
import os
import pathlib
from os.path import exists
from urllib.parse import urlparse

import pandas as pd
import smart_open
from airflow.hooks.base import BaseHook

from universal_transfer_operator.constants import FileLocation
from universal_transfer_operator.data_providers.base import DataStream
from universal_transfer_operator.data_providers.filesystem.base import BaseFilesystemProviders


class LocalDataProvider(BaseFilesystemProviders):
    """Handler Local file path operations"""

    location_type = FileLocation.LOCAL

    @property
    def paths(self) -> list[str]:
        """Resolve local filepath"""
        url = urlparse(self.dataset.path)
        path_object = pathlib.Path(url.path)
        if path_object.is_dir():
            paths = [str(filepath) for filepath in path_object.rglob("*")]
        else:
            paths = glob.glob(url.path)
        return paths

    def validate_conn(self):
        """Override as conn_id is not always required for local location."""

    @property
    def size(self) -> int:
        """Return the size in bytes of the given file.
        :return: File size in bytes
        """
        path = pathlib.Path(self.dataset.path)
        return os.path.getsize(path)

    @property
    def openlineage_dataset_namespace(self) -> str:
        """
        Returns the open lineage dataset namespace as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        return str(os.path.basename(self.dataset.path))

    @property
    def openlineage_dataset_name(self) -> str:
        """
        Returns the open lineage dataset name as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        return str(urlparse(self.dataset.path).path)

    def delete(self, path: str | None = None):
        """
        Delete a file/object if they exists
        """
        path = self.dataset.path if path is None else path
        os.remove(path)

    def check_if_exists(self, path: str | None = None) -> bool:
        """Return true if the dataset exists"""
        path = self.dataset.path if path is None else path
        return exists(path)

    def write_using_smart_open(self, source_ref: DataStream | pd.DataFrame) -> str:
        """Write the source data from remote object i/o buffer to the dataset using smart open"""
        mode = "wb" if self.read_as_binary(self.dataset.path) else "w"
        with smart_open.open(self.dataset.path, mode=mode, transport_params=self.transport_params) as stream:
            if isinstance(source_ref, DataStream):
                stream.write(source_ref.remote_obj_buffer.read())
            elif isinstance(source_ref, pd.DataFrame):
                self.dataset.type.create_from_dataframe(stream=stream, df=source_ref)
        return self.dataset.path

    @property
    def hook(self) -> BaseHook:
        """Return an instance of the Airflow hook."""
        raise NotImplementedError



from __future__ import annotations

from attr import define

from universal_transfer_operator.datasets.base import Dataset


@define
class API(Dataset):
    """
    Repersents all API dataset.
    Intended to be used within library.

    :param path: Path to a file in the filesystem/Object stores
    """

    # TODO: define the name and namespace for API



from __future__ import annotations

import os
from contextlib import contextmanager
from functools import cached_property
from tempfile import NamedTemporaryFile
from typing import Any, Iterator
from urllib.parse import urlparse, urlunparse

from airflow.providers.amazon.aws.hooks.s3 import S3Hook

from universal_transfer_operator.constants import FileLocation, Location, TransferMode
from universal_transfer_operator.data_providers.filesystem.base import (
    BaseFilesystemProviders,
    Path,
    TempFile,
)
from universal_transfer_operator.datasets.file.base import File
from universal_transfer_operator.exceptions import DatabaseCustomError
from universal_transfer_operator.integrations.base import TransferIntegrationOptions


class S3DataProvider(BaseFilesystemProviders):
    """
    DataProviders interactions with S3 Dataset.
    """

    location_type = FileLocation.S3

    def __init__(
        self,
        dataset: File,
        transfer_params: TransferIntegrationOptions = TransferIntegrationOptions(),
        transfer_mode: TransferMode = TransferMode.NONNATIVE,
    ):
        super().__init__(
            dataset=dataset,
            transfer_params=transfer_params,
            transfer_mode=transfer_mode,
        )
        self.transfer_mapping: set = {
            Location.S3,
            Location.GS,
        }

    @cached_property
    def hook(self) -> S3Hook:
        """Return an instance of the database-specific Airflow hook."""
        return S3Hook(
            aws_conn_id=self.dataset.conn_id,
            verify=self.verify,
            transfer_config_args=self.transfer_config_args,
            extra_args=self.s3_extra_args,
        )

    def delete(self, path: str | None = None):
        """
        Delete a file/object if they exists
        """
        path = self.dataset.path if path is None else path
        url = urlparse(path)
        self.hook.delete_objects(bucket=url.netloc, keys=url.path.lstrip("/"))

    @property
    def transport_params(self) -> dict:
        """Structure s3fs credentials from Airflow connection.
        s3fs enables pandas to write to s3
        """
        return {"client": self.hook.conn}

    @property
    def paths(self) -> list[str]:
        """Resolve S3 file paths with prefix"""
        url = urlparse(self.dataset.path)
        prefixes = self.hook.list_keys(
            bucket_name=self.bucket_name,
            prefix=self.s3_key,
            delimiter=self.delimiter,
        )
        paths = [urlunparse((url.scheme, url.netloc, keys, "", "", "")) for keys in prefixes]
        return paths

    def check_if_exists(self, path: str | None = None) -> Any:
        """Return true if the dataset exists"""
        path = self.dataset.path if path is None else path
        return self.hook.check_for_key(key=path)

    @contextmanager
    def read_using_hook(self) -> Iterator[list[TempFile]]:
        """Read the file from dataset and write to local file location"""
        if not self.check_if_exists():
            raise ValueError(f"{self.dataset.path} doesn't exits")
        files = self.hook.list_keys(
            bucket_name=self.bucket_name,
            prefix=self.s3_key,
            delimiter=self.delimiter,
        )
        local_file_paths = []
        try:
            for file in files:
                local_file_paths.append(self.download_file(file))
            yield local_file_paths
        finally:
            # Clean up the local files
            self.cleanup(local_file_paths)

    def write_using_hook(self, source_ref: list[TempFile]):
        """Write the file from local file location to the dataset"""

        dest_s3_key = self.dataset.path

        if not self.keep_directory_structure and self.prefix:
            dest_s3_key = os.path.join(dest_s3_key, self.prefix)

        destination_keys = []
        for file in source_ref:
            if file.tmp_file and file.tmp_file.exists():
                dest_key = os.path.join(dest_s3_key, os.path.basename(file.actual_filename.name))
                self.hook.load_file(
                    filename=file.tmp_file.as_posix(),
                    key=dest_key,
                    replace=True,
                    acl_policy=self.s3_acl_policy,
                )
                destination_keys.append(dest_key)

        return destination_keys

    def download_file(self, file) -> TempFile:
        """Download file and save to temporary path."""
        file_object = self.hook.get_key(file, self.bucket_name)
        _, _, file_name = file.rpartition("/")
        with NamedTemporaryFile(suffix=file_name, delete=False) as tmp_file:
            file_object.download_fileobj(tmp_file)
            return TempFile(tmp_file=Path(tmp_file.name), actual_filename=Path(file_name))

    @property
    def verify(self) -> Any:
        return self.dataset.extra.get("verify")

    @property
    def transfer_config_args(self) -> Any:
        return self.dataset.extra.get("transfer_config_args", {})

    @property
    def s3_extra_args(self) -> Any:
        return self.dataset.extra.get("s3_extra_args", {})

    @property
    def bucket_name(self) -> Any:
        bucket_name, _ = self.hook.parse_s3_url(self.dataset.path)
        return bucket_name

    @property
    def s3_key(self) -> Any:
        _, key = self.hook.parse_s3_url(self.dataset.path)
        return key

    @property
    def s3_acl_policy(self) -> Any:
        return self.dataset.extra.get("s3_acl_policy", None)

    @property
    def prefix(self) -> Any:
        return self.dataset.extra.get("prefix", None)

    @property
    def keep_directory_structure(self) -> Any:
        return self.dataset.extra.get("keep_directory_structure", False)

    @property
    def delimiter(self) -> Any:
        return self.dataset.extra.get("delimiter", None)

    @property
    def openlineage_dataset_namespace(self) -> str:
        """
        Returns the open lineage dataset namespace as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        raise NotImplementedError

    @property
    def openlineage_dataset_name(self) -> str:
        """
        Returns the open lineage dataset name as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        raise NotImplementedError

    @property
    def openlineage_dataset_uri(self) -> str:
        """
        Returns the open lineage dataset uri as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        raise NotImplementedError

    @property
    def size(self) -> int:
        """Return file size for S3 location"""
        url = urlparse(self.dataset.path)
        bucket_name = url.netloc
        object_name = url.path
        if object_name.startswith("/"):
            object_name = object_name[1:]
        obj = self.hook.head_object(key=object_name, bucket_name=bucket_name)
        return obj.get("ContentLength") if obj else -1  # type: ignore

    def get_snowflake_stage_auth_sub_statement(self) -> str:
        aws = self.hook.get_credentials()
        if aws.access_key and aws.secret_key:
            auth = f"credentials=(aws_key_id='{aws.access_key}' aws_secret_key='{aws.secret_key}');"
        else:
            raise DatabaseCustomError(
                "In order to create an stage for S3, one of the following is required: "
                "* `storage_integration`"
                "* AWS_KEY_ID and SECRET_KEY_ID"
            )
        return auth



from __future__ import annotations

import io
from abc import ABC
from pathlib import Path
from typing import Generic, Iterator, TypeVar

import attr
import pandas as pd

from universal_transfer_operator.datasets.dataframe.base import Dataframe
from universal_transfer_operator.datasets.file.base import File
from universal_transfer_operator.datasets.table import Table

DatasetType = TypeVar("DatasetType", File, Table, Dataframe)


@attr.define
class DataStream:
    remote_obj_buffer: io.IOBase
    actual_filename: Path
    actual_file: File


class DataProviders(ABC, Generic[DatasetType]):
    """
    Base class to represent all the DataProviders interactions with Dataset.

    The goal is to be able to support new dataset by adding
    a new module to the `uto/data_providers` directory, without the need of
    changing other modules and classes.
    """

    def __init__(
        self,
        dataset: DatasetType,
    ):
        self.dataset: DatasetType = dataset

    def read(self) -> Iterator[pd.DataFrame] | Iterator[DataStream]:
        """Read from filesystem dataset or databases dataset and write to local reference locations or dataframes"""
        raise NotImplementedError

    def write(self, source_ref: pd.DataFrame | DataStream) -> str | DataProviders:  # type: ignore
        """Write the data from local reference location or a dataframe to the database dataset or filesystem dataset

        :param source_ref: Stream of data to be loaded into output table or a pandas dataframe.
        """
        raise NotImplementedError



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Iterator

import pandas as pd
import sqlalchemy

if TYPE_CHECKING:  # pragma: no cover
    from sqlalchemy.engine.cursor import CursorResult

import logging
import warnings

from airflow.hooks.dbapi import DbApiHook
from pandas.io.sql import SQLDatabase
from sqlalchemy.sql import ClauseElement
from sqlalchemy.sql.schema import Table as SqlaTable

from universal_transfer_operator.constants import (
    DEFAULT_CHUNK_SIZE,
    LoadExistStrategy,
    Location,
    TransferMode,
)
from universal_transfer_operator.data_providers.base import DataProviders, DataStream
from universal_transfer_operator.data_providers.filesystem import resolve_file_path_pattern
from universal_transfer_operator.datasets.file.base import File
from universal_transfer_operator.datasets.table import Metadata, Table
from universal_transfer_operator.settings import (
    LOAD_TABLE_AUTODETECT_ROWS_COUNT,
    SCHEMA,
)
from universal_transfer_operator.universal_transfer_operator import TransferIntegrationOptions
from universal_transfer_operator.utils import get_dataset_connection_type


class DatabaseDataProvider(DataProviders[Table]):
    """DatabaseProviders represent all the DataProviders interactions with Databases."""

    _create_schema_statement: str = "CREATE SCHEMA IF NOT EXISTS {}"
    _drop_table_statement: str = "DROP TABLE IF EXISTS {}"
    _create_table_statement: str = "CREATE TABLE IF NOT EXISTS {} AS {}"
    # Used to normalize the ndjson when appending fields in nested fields.
    # Example -
    #   ndjson - {'a': {'b': 'val'}}
    #   the col names generated is 'a.b'. char '.' maybe an illegal char in some db's col name.
    # Contains the illegal char and there replacement, where the value in
    # illegal_column_name_chars[0] will be replaced by value in illegal_column_name_chars_replacement[0]
    illegal_column_name_chars: list[str] = []
    illegal_column_name_chars_replacement: list[str] = []
    # In run_raw_sql operator decides if we want to return results directly or process them by handler provided
    IGNORE_HANDLER_IN_RUN_RAW_SQL: bool = False
    NATIVE_PATHS: dict[Any, Any] = {}
    DEFAULT_SCHEMA = SCHEMA

    def __init__(
        self,
        dataset: Table,
        transfer_mode: TransferMode,
        transfer_params: TransferIntegrationOptions = TransferIntegrationOptions(),
    ):
        self.dataset = dataset
        self.transfer_params = transfer_params
        self.if_exists = self._if_exists
        self.transfer_mode = transfer_mode
        self.transfer_mapping: set[Location] = set()
        self.LOAD_DATA_NATIVELY_FROM_SOURCE: dict = {}
        super().__init__(dataset=self.dataset)

    def __repr__(self):
        return f'{self.__class__.__name__}(conn_id="{self.dataset.conn_id})'

    @property
    def sql_type(self):
        raise NotImplementedError

    @property
    def hook(self) -> DbApiHook:
        """Return an instance of the database-specific Airflow hook."""
        raise NotImplementedError

    @property
    def connection(self) -> sqlalchemy.engine.base.Connection:
        """Return a Sqlalchemy connection object for the given database."""
        return self.sqlalchemy_engine.connect()

    @property
    def sqlalchemy_engine(self) -> sqlalchemy.engine.base.Engine:
        """Return Sqlalchemy engine."""
        return self.hook.get_sqlalchemy_engine()  # type: ignore[no-any-return]

    @property
    def transport_params(self) -> dict | None:  # skipcq: PYL-R0201
        """Get credentials required by smart open to access files"""
        return None

    @property
    def _if_exists(self) -> LoadExistStrategy:
        """Check if_exists is passed by user else default to replace"""
        try:
            if_exist = self.transfer_params.if_exists
        except AttributeError:
            if_exist = "replace"
        return if_exist

    def run_sql(
        self,
        sql: str | ClauseElement = "",
        parameters: dict | None = None,
        handler: Callable | None = None,
        **kwargs,
    ) -> CursorResult:
        """
        Return the results to running a SQL statement.

        Whenever possible, this method should be implemented using Airflow Hooks,
        since this will simplify the integration with Async operators.

        :param sql: Contains SQL query to be run against database
        :param parameters: Optional parameters to be used to render the query
        :param autocommit: Optional autocommit flag
        """
        if parameters is None:
            parameters = {}

        if "sql_statement" in kwargs:  # pragma: no cover
            warnings.warn(
                "`sql_statement` is deprecated and will be removed in future release"
                "Please use  `sql` param instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            sql = kwargs.get("sql_statement")  # type: ignore

        # We need to autocommit=True to make sure the query runs. This is done exclusively for SnowflakeDatabase's
        # truncate method to reflect changes.
        if isinstance(sql, str):
            result = self.connection.execute(
                sqlalchemy.text(sql).execution_options(autocommit=True), parameters
            )
        else:
            result = self.connection.execute(sql, parameters)
        if handler:
            return handler(result)
        return None

    def columns_exist(self, table: Table, columns: list[str]) -> bool:
        """
        Check that a list of columns exist in the given table.

        :param table: The table to check in.
        :param columns: The columns to check.

        :returns: whether the columns exist in the table or not.
        """
        sqla_table = self.get_sqla_table(table)
        return all(
            any(sqla_column.name == column for sqla_column in sqla_table.columns) for column in columns
        )

    def get_sqla_table(self, table: Table) -> SqlaTable:
        """
        Return SQLAlchemy table instance

        :param table: Astro Table to be converted to SQLAlchemy table instance
        """
        return SqlaTable(
            table.name,
            table.sqlalchemy_metadata,
            autoload_with=self.sqlalchemy_engine,
            extend_existing=True,
        )

    def table_exists(self, table: Table) -> bool:
        """
        Check if a table exists in the database.

        :param table: Details of the table we want to check that exists
        """
        table_qualified_name = self.get_table_qualified_name(table)
        inspector = sqlalchemy.inspect(self.sqlalchemy_engine)
        return bool(inspector.dialect.has_table(self.connection, table_qualified_name))

    def check_if_transfer_supported(self, source_dataset: Table) -> bool:
        """
        Checks if the transfer is supported from source to destination based on source_dataset.

        :param source_dataset: Table present in the source location
        """
        source_connection_type = get_dataset_connection_type(source_dataset)
        return Location(source_connection_type) in self.transfer_mapping

    def read(self) -> Iterator[pd.DataFrame]:
        """Convert a Table into a Pandas DataFrame"""
        if self.transfer_mode == TransferMode.NATIVE:
            raise ValueError("No native path for `database` to destination")
        else:
            yield self.export_table_to_pandas_dataframe()

    def write(self, source_ref: DataStream | pd.DataFrame) -> str:
        """
        Write the data from local reference location or dataframe to the database dataset or filesystem dataset.

        :param source_ref: Stream of data to be loaded into output table or a pandas dataframe.
        """
        # `source_ref` can be a dataframe for all the filetypes we can create a dataframe for like -
        # CSV, JSON, NDJSON, and Parquet or SQL Tables. This gives us the option to perform various
        # functions on the data on the fly, like filtering or changing the file format altogether. For other
        # files whose content cannot be converted to dataframe like - zip or image, we get a DataStream object.
        if isinstance(source_ref, DataStream):
            return self.load_file_to_table(
                input_file=source_ref.actual_file, output_table=self.dataset, if_exists=self.if_exists
            )
        return self.load_dataframe_to_table(
            input_dataframe=source_ref, output_table=self.dataset, if_exists=self.if_exists
        )

    def load_dataframe_to_table(
        self,
        input_dataframe: pd.DataFrame,
        output_table: Table,
        if_exists: LoadExistStrategy = "replace",
        chunk_size: int = DEFAULT_CHUNK_SIZE,
    ) -> str:
        """
        Load content of dataframe in output_table.

        :param input_dataframe: dataframe
        :param output_table: Table to create
        :param if_exists: Overwrite file if exists
        :param chunk_size: Specify the number of records in each batch to be written at a time
        :param normalize_config: pandas json_normalize params config
        """

        self.create_schema_and_table_if_needed_from_dataframe(
            table=output_table,
            dataframe=input_dataframe,
            if_exists=if_exists,
        )
        self.load_pandas_dataframe_to_table(
            input_dataframe,
            output_table,
            chunk_size=chunk_size,
            if_exists=if_exists,
        )
        return self.get_table_qualified_name(output_table)

    @property
    def openlineage_dataset_namespace(self) -> str:
        """
        Returns the open lineage dataset namespace as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        raise NotImplementedError

    @property
    def openlineage_dataset_name(self) -> str:
        """
        Returns the open lineage dataset name as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        raise NotImplementedError

    @property
    def openlineage_dataset_uri(self) -> str:
        """
        Returns the open lineage dataset uri as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        raise NotImplementedError

    # ---------------------------------------------------------
    # Table metadata
    # ---------------------------------------------------------
    @staticmethod
    def get_table_qualified_name(table: Table) -> str:  # skipcq: PYL-R0201
        """
        Return table qualified name. This is Database-specific.
        For instance, in Sqlite this is the table name. In Snowflake, however, it is the database, schema and table

        :param table: The table we want to retrieve the qualified name for.
        """
        # Initially this method belonged to the Table class.
        # However, in order to have an agnostic table class implementation,
        # we are keeping all methods which vary depending on the database within the Database class.
        if table.metadata and table.metadata.schema:
            qualified_name = f"{table.metadata.schema}.{table.name}"
        else:
            qualified_name = table.name
        return qualified_name

    @property
    def default_metadata(self) -> Metadata:
        """
        Extract the metadata available within the Airflow connection associated with
        self.dataset.conn_id.

        :return: a Metadata instance
        """
        raise NotImplementedError

    def populate_metadata(self):
        """
        Given a table, check if the table has metadata.
        If the metadata is missing, and the database has metadata, assign it to the table.
        If the table schema was not defined by the end, retrieve the user-defined schema.
        This method performs the changes in-place and also returns the table.

        :param table: Table to potentially have their metadata changed
        :return table: Return the modified table
        """

        if self.dataset.metadata and self.dataset.metadata.is_empty() and self.default_metadata:
            self.dataset.metadata = self.default_metadata
        if not self.dataset.metadata.schema:
            self.dataset.metadata.schema = self.DEFAULT_SCHEMA

    # ---------------------------------------------------------
    # Table creation & deletion methods
    # ---------------------------------------------------------
    def create_table_using_columns(self, table: Table) -> None:
        """
        Create a SQL table using the table columns.

        :param table: The table to be created.
        """
        if not table.columns:
            raise ValueError("To use this method, table.columns must be defined")

        metadata = table.sqlalchemy_metadata
        sqlalchemy_table = sqlalchemy.Table(table.name, metadata, *table.columns)
        metadata.create_all(self.sqlalchemy_engine, tables=[sqlalchemy_table])

    def is_native_autodetect_schema_available(  # skipcq: PYL-R0201
        self, file: File  # skipcq: PYL-W0613
    ) -> bool:
        """
        Check if native auto detection of schema is available.

        :param file: File used to check the file type of to decide
            whether there is a native auto detection available for it.
        """
        return False

    def create_table_using_native_schema_autodetection(
        self,
        table: Table,
        file: File,
    ) -> None:
        """
        Create a SQL table, automatically inferring the schema using the given file via native database support.

        :param table: The table to be created.
        :param file: File used to infer the new table columns.
        """
        raise NotImplementedError("Missing implementation of native schema autodetection.")

    def create_table_using_schema_autodetection(
        self,
        table: Table,
        file: File | None = None,
        dataframe: pd.DataFrame | None = None,
    ) -> None:
        """
        Create a SQL table, automatically inferring the schema using the given file.

        :param table: The table to be created.
        :param file: File used to infer the new table columns.
        :param dataframe: Dataframe used to infer the new table columns if there is no file
        """
        if file is None:
            if dataframe is None:
                raise ValueError(
                    "File or Dataframe is required for creating table using schema autodetection"
                )
            source_dataframe = dataframe
        else:
            source_dataframe = file.export_to_dataframe(nrows=LOAD_TABLE_AUTODETECT_ROWS_COUNT)

        db = SQLDatabase(engine=self.sqlalchemy_engine)

        db.prep_table(
            source_dataframe,
            table.name.lower(),
            schema=table.metadata.schema,
            if_exists="replace",
            index=False,
        )

    def create_table(
        self,
        table: Table,
        file: File | None = None,
        dataframe: pd.DataFrame | None = None,
        use_native_support: bool = True,
    ) -> None:
        """
        Create a table either using its explicitly defined columns or inferring
        it's columns from a given file.

        :param table: The table to be created
        :param file: (optional) File used to infer the table columns.
        :param dataframe: (optional) Dataframe used to infer the new table columns if there is no file
        """
        if table.columns:
            self.create_table_using_columns(table)
        elif use_native_support and file and self.is_native_autodetect_schema_available(file):
            self.create_table_using_native_schema_autodetection(table, file)
        else:
            self.create_table_using_schema_autodetection(
                table,
                file=file,
                dataframe=dataframe,
            )

    def create_table_from_select_statement(
        self,
        statement: str,
        target_table: Table,
        parameters: dict | None = None,
    ) -> None:
        """
        Export the result rows of a query statement into another table.

        :param statement: SQL query statement
        :param target_table: Destination table where results will be recorded.
        :param parameters: (Optional) parameters to be used to render the SQL query
        """
        statement = self._create_table_statement.format(
            self.get_table_qualified_name(target_table), statement
        )
        self.run_sql(statement, parameters)

    def drop_table(self, table: Table) -> None:
        """
        Delete a SQL table, if it exists.

        :param table: The table to be deleted.
        """
        statement = self._drop_table_statement.format(self.get_table_qualified_name(table))
        self.run_sql(statement)

    # ---------------------------------------------------------
    # Table load methods
    # ---------------------------------------------------------

    def create_schema_and_table_if_needed(
        self,
        table: Table,
        file: File,
        normalize_config: dict | None = None,
        if_exists: LoadExistStrategy = "replace",
        use_native_support: bool = True,
    ):
        """
        Checks if the autodetect schema exists for native support else creates the schema and table
        :param table: Table to create
        :param file: File path and conn_id for object stores
        :param normalize_config: pandas json_normalize params config
        :param if_exists:  Overwrite file if exists
        :param use_native_support: Use native support for data transfer if available on the destination
        """
        is_schema_autodetection_supported = self.check_schema_autodetection_is_supported(source_file=file)
        is_file_pattern_based_schema_autodetection_supported = (
            self.check_file_pattern_based_schema_autodetection_is_supported(source_file=file)
        )
        if if_exists == "replace":
            self.drop_table(table)
        if use_native_support and is_schema_autodetection_supported and not file.is_pattern():
            return
        if (
            use_native_support
            and file.is_pattern()
            and is_schema_autodetection_supported
            and is_file_pattern_based_schema_autodetection_supported
        ):
            return

        self.create_schema_if_needed(table.metadata.schema)
        if if_exists == "replace" or not self.table_exists(table):
            files = resolve_file_path_pattern(
                file,
                normalize_config=normalize_config,
                filetype=file.type.name,
                transfer_params=self.transfer_params,
                transfer_mode=self.transfer_mode,
            )
            self.create_table(
                table,
                # We only use the first file for inferring the table schema
                files[0],
                use_native_support=use_native_support,
            )

    def create_schema_and_table_if_needed_from_dataframe(
        self,
        table: Table,
        dataframe: pd.DataFrame,
        if_exists: LoadExistStrategy = "replace",
        use_native_support: bool = True,
    ):
        """
        Creates the schema and table from dataframe

        :param table: Table to create
        :param dataframe: dataframe object to be used as a source of data
        :param if_exists:  Overwrite file if exists
        :param use_native_support: Use native support for data transfer if available on the destination
        """
        if if_exists == "replace":
            self.drop_table(table)
        self.create_schema_if_needed(table.metadata.schema)
        if if_exists == "replace" or not self.table_exists(table):
            self.create_table(
                table,
                dataframe=dataframe,
                use_native_support=use_native_support,
            )

    def fetch_all_rows(self, table: Table, row_limit: int = -1) -> list:
        """
        Fetches all rows for a table and returns as a list. This is needed because some
        databases have different cursors that require different methods to fetch rows

        :param row_limit: Limit the number of rows returned, by default return all rows.
        :param table: The table metadata needed to fetch the rows
        :return: a list of rows
        """
        statement = f"SELECT * FROM {self.get_table_qualified_name(table)}"  # skipcq: BAN-B608
        if row_limit > -1:
            statement = statement + f" LIMIT {row_limit}"  # skipcq: BAN-B608
        response = self.run_sql(statement, handler=lambda x: x.fetchall())
        return response  # type: ignore

    def load_file_to_table(
        self,
        input_file: File,
        output_table: Table,
        normalize_config: dict | None = None,
        if_exists: LoadExistStrategy = "replace",
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        **kwargs,
    ) -> str:
        """
        Load content of multiple files in output_table.
        Multiple files are sourced from the file path, which can also be path pattern.

        :param input_file: File path and conn_id for object stores
        :param output_table: Table to create
        :param if_exists: Overwrite file if exists
        :param chunk_size: Specify the number of records in each batch to be written at a time
        :param use_native_support: Use native support for data transfer if available on the destination
        :param normalize_config: pandas json_normalize params config
        """
        normalize_config = normalize_config or {}

        self.create_schema_and_table_if_needed(
            file=input_file,
            table=output_table,
            if_exists=if_exists,
            normalize_config=normalize_config,
        )
        if self.transfer_mode == TransferMode.NATIVE:
            logging.info("Loading file(s) with Native Support...")
            self.load_file_to_table_natively(
                source_file=input_file,
                target_table=output_table,
                if_exists="append",
                **kwargs,
            )
        else:
            logging.info("Loading file(s) with Non Native Support...")
            self.load_file_to_table_using_pandas(
                input_file=input_file,
                output_table=output_table,
                normalize_config=normalize_config,
                if_exists="append",
                chunk_size=chunk_size,
            )
        return self.get_table_qualified_name(output_table)

    def load_file_to_table_natively(
        self,
        source_file: File,
        target_table: Table,
        if_exists: LoadExistStrategy = "replace",
        **kwargs,
    ):
        """
        Checks if optimised path for transfer between File location to database exists
        and if it does, it transfers it and returns true else false

        :param source_file: File from which we need to transfer data
        :param target_table: Table that needs to be populated with file data
        :param if_exists: Overwrite file if exists. Default False
        """
        raise NotImplementedError

    def load_file_to_table_using_pandas(
        self,
        input_file: File,
        output_table: Table,
        normalize_config: dict | None = None,
        if_exists: LoadExistStrategy = "replace",
        chunk_size: int = DEFAULT_CHUNK_SIZE,
    ):
        input_files = resolve_file_path_pattern(
            file=input_file,
            normalize_config=normalize_config,
            filetype=input_file.type.name,
            transfer_params=self.transfer_params,
            transfer_mode=self.transfer_mode,
        )

        for file in input_files:
            self.load_pandas_dataframe_to_table(
                self.get_dataframe_from_file(file),
                output_table,
                chunk_size=chunk_size,
                if_exists=if_exists,
            )

    def load_pandas_dataframe_to_table(
        self,
        source_dataframe: pd.DataFrame,
        target_table: Table,
        if_exists: LoadExistStrategy = "replace",
        chunk_size: int = DEFAULT_CHUNK_SIZE,
    ) -> None:
        """
        Create a table with the dataframe's contents.
        If the table already exists, append or replace the content, depending on the value of `if_exists`.

        :param source_dataframe: Local or remote filepath
        :param target_table: Table in which the file will be loaded
        :param if_exists: Strategy to be used in case the target table already exists.
        :param chunk_size: Specify the number of rows in each batch to be written at a time.
        """
        self._assert_not_empty_df(source_dataframe)

        source_dataframe.to_sql(
            self.get_table_qualified_name(target_table),
            con=self.sqlalchemy_engine,
            if_exists=if_exists,
            chunksize=chunk_size,
            method="multi",
            index=False,
        )

    @staticmethod
    def _assert_not_empty_df(df: pd.DataFrame) -> None:
        """Raise error if dataframe empty

        param df: A dataframe
        """
        if df.empty:
            raise ValueError("Can't load empty dataframe")

    @staticmethod
    def get_dataframe_from_file(file: File):
        """
        Get pandas dataframe file. We need export_to_dataframe() for Biqqery,Snowflake and Redshift except for Postgres.
        For postgres we are overriding this method and using export_to_dataframe_via_byte_stream().
        export_to_dataframe_via_byte_stream copies files in a buffer and then use that buffer to ingest data.
        With this approach we have significant performance boost for postgres.

        :param file: File path and conn_id for object stores
        """

        return file.export_to_dataframe()

    def check_schema_autodetection_is_supported(  # skipcq: PYL-R0201
        self, source_file: File  # skipcq: PYL-W0613
    ) -> bool:
        """
        Checks if schema autodetection is handled natively by the database. Return False by default.

        :param source_file: File from which we need to transfer data
        """
        return False

    def check_file_pattern_based_schema_autodetection_is_supported(  # skipcq: PYL-R0201
        self, source_file: File  # skipcq: PYL-W0613
    ) -> bool:
        """
        Checks if schema autodetection is handled natively by the database for file
        patterns and prefixes. Return False by default.

        :param source_file: File from which we need to transfer data
        """
        return False

    def row_count(self, table: Table):
        """
        Returns the number of rows in a table.

        :param table: table to count
        :return: The number of rows in the table
        """
        result = self.run_sql(
            f"select count(*) from {self.get_table_qualified_name(table)}",
            handler=lambda x: x.scalar(),  # skipcq: BAN-B608
        )
        return result

    # ---------------------------------------------------------
    # Schema Management
    # ---------------------------------------------------------

    def create_schema_if_needed(self, schema: str | None) -> None:
        """
        This function checks if the expected schema exists in the database. If the schema does not exist,
        it will attempt to create it.

        :param schema: DB Schema - a namespace that contains named objects like (tables, functions, etc)
        """
        # We check if the schema exists first because snowflake will fail on a create schema query even if it
        # doesn't actually create a schema.
        if schema and not self.schema_exists(schema):
            statement = self._create_schema_statement.format(schema)
            self.run_sql(statement)

    def schema_exists(self, schema: str) -> bool:
        """
        Checks if a schema exists in the database

        :param schema: DB Schema - a namespace that contains named objects like (tables, functions, etc)
        """
        raise NotImplementedError

    # ---------------------------------------------------------
    # Extract methods
    # ---------------------------------------------------------

    def export_table_to_pandas_dataframe(self) -> pd.DataFrame:
        """
        Copy the content of a table to an in-memory Pandas dataframe.
        """

        if not self.table_exists(self.dataset):
            raise ValueError(f"The table {self.dataset.name} does not exist")

        sqla_table = self.get_sqla_table(self.dataset)
        return pd.read_sql(sql=sqla_table.select(), con=self.sqlalchemy_engine)

    def is_native_path_available(
        self,
        source_dataset: Table,  # skipcq PYL-W0613, PYL-R0201
    ) -> bool:
        """
        Check if there is an optimised path for source to destination.

        :param source_dataset: Dataframe from which we need to transfer data
        """
        return False



from __future__ import annotations

try:
    # Airflow >= 2.4
    from airflow.datasets import Dataset

    DATASET_SUPPORT = True
except ImportError:
    # Airflow < 2.4
    Dataset = object  # type: ignore
    DATASET_SUPPORT = False



from __future__ import annotations

import sys
from enum import Enum

# typing.Literal was only introduced in Python 3.8, and we support Python 3.7
if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class Location(Enum):
    LOCAL = "local"
    GS = "gs"  # Google Cloud Storage
    S3 = "s3"  # Amazon S3
    SFTP = "sftp"
    SQLITE = "sqlite"
    BIGQUERY = "bigquery"
    SNOWFLAKE = "snowflake"

    def __repr__(self):
        return f"{self}"


class FileLocation(Enum):
    # [START filelocation]
    LOCAL = "local"
    GS = "gs"  # Google Cloud Storage
    S3 = "s3"  # Amazon S3
    SFTP = "sftp"
    # [END filelocation]

    def __repr__(self):
        return f"{self}"


class IngestorSupported(Enum):
    # [START transferingestor]
    Fivetran = "fivetran"
    # [END transferingestor]

    def __repr__(self):
        return f"{self}"


class TransferMode(Enum):
    # [START TransferMode]
    NATIVE = "native"
    NONNATIVE = "nonnative"
    THIRDPARTY = "thirdparty"
    # [END TransferMode]

    def __str__(self) -> str:
        return self.value


class FileType(Enum):
    # [START filetypes]
    CSV = "csv"
    JSON = "json"
    NDJSON = "ndjson"
    PARQUET = "parquet"
    # [END filetypes]

    def __repr__(self):
        return f"{self}"


class Database(Enum):
    # [START database]
    SQLITE = "sqlite"
    BIGQUERY = "bigquery"
    SNOWFLAKE = "snowflake"
    # [END database]

    def __repr__(self):
        return f"{self}"


class FivetranConnectorSupported(Enum):
    # [START FivetranConnectorSupported]
    S3 = "S3"
    # [END FivetranConnectorSupported]

    def __str__(self) -> str:
        return self.value


class FivetranDestinationSupported(Enum):
    # [START FivetranDestinationSupported]
    SNOWFLAKE = "snowflake"
    # [END FivetranDestinationSupported]

    def __str__(self) -> str:
        return self.value


SUPPORTED_FILE_LOCATIONS = [const.value for const in FileLocation]
SUPPORTED_FILE_TYPES = [const.value for const in FileType]
SUPPORTED_DATABASES = [const.value for const in Database]

# [START LoadExistStrategy]
LoadExistStrategy = Literal["replace", "append"]
# [END LoadExistStrategy]

DEFAULT_CHUNK_SIZE = 1000000
ColumnCapitalization = Literal["upper", "lower", "original"]
DEFAULT_SCHEMA = "tmp_transfers"
IAM_ROLE_ACTIVATION_WAIT_TIME = 60



from __future__ import annotations

import os
from functools import cached_property
from urllib.parse import ParseResult, urlparse, urlunparse

import pandas as pd
import smart_open
from airflow.providers.sftp.hooks.sftp import SFTPHook

from universal_transfer_operator.constants import FileLocation, Location, TransferMode
from universal_transfer_operator.data_providers.base import DataStream
from universal_transfer_operator.data_providers.filesystem.base import BaseFilesystemProviders
from universal_transfer_operator.datasets.file.base import File
from universal_transfer_operator.integrations.base import TransferIntegrationOptions


class SFTPDataProvider(BaseFilesystemProviders):
    """
    DataProviders interactions with GS Dataset.
    """

    location_type = FileLocation.SFTP

    def __init__(
        self,
        dataset: File,
        transfer_params: TransferIntegrationOptions = TransferIntegrationOptions(),
        transfer_mode: TransferMode = TransferMode.NONNATIVE,
    ):
        super().__init__(
            dataset=dataset,
            transfer_params=transfer_params,
            transfer_mode=transfer_mode,
        )
        self.transfer_mapping = {
            Location.S3,
            Location.GS,
        }

    @cached_property
    def hook(self) -> SFTPHook:
        """Return an instance of the SFTPHook Airflow hook."""
        return SFTPHook(ssh_conn_id=self.dataset.conn_id)

    def delete(self, path: str | None = None):
        """
        Delete a file/object if they exists
        """
        path = self.dataset.path if path is None else path
        self.hook.delete_file(path=path.replace("sftp://", "/"))

    def check_if_exists(self, path: str | None = None):
        """Return true if the dataset exists"""
        path = self.dataset.path if path is None else path
        return self.hook.path_exists(path.replace("sftp://", "/"))

    @property
    def paths(self) -> list[str]:
        """
        Resolve SFTP file paths with netloc of self.dataset.path as prefix. Paths are added if they start with prefix

        Example - if there are multiple paths like
            - sftp://upload/test.csv
            - sftp://upload/test.json
            - sftp://upload/home.parquet
            - sftp://upload/sample.ndjson

        If self.dataset.path is "sftp://upload/test" will return sftp://upload/test.csv and sftp://upload/test.json
        """
        url = urlparse(self.dataset.path)
        uri = self.get_uri()
        full_paths = []
        prefixes = self.hook.get_tree_map(url.netloc, prefix=url.netloc + url.path)
        for keys in prefixes:
            if len(keys) > 0:
                full_paths.extend(keys)
        # paths = ["/" + path for path in full_paths]
        paths = [uri + "/" + path for path in full_paths]
        return paths

    @property
    def transport_params(self) -> dict:
        """get SFTP credentials for storage"""
        client = self.hook.get_connection(self.dataset.conn_id)
        extra_options = client.extra_dejson
        if "key_file" in extra_options:
            key_file = extra_options.get("key_file")
            return {"connect_kwargs": {"key_filename": key_file}}
        elif client.password:
            return {"connect_kwargs": {"password": client.password}}
        raise ValueError("SFTP credentials are not set in the connection.")

    def get_uri(self):
        client = self.hook.get_connection(self.dataset.conn_id)
        return client.get_uri()

    @staticmethod
    def _get_url_path(dst_url: ParseResult, src_url: ParseResult) -> str:
        """
        Get correct file path, priority is given to destination file path.
        :return: URL path
        """
        path = dst_url.path if dst_url.__getattribute__("path") else src_url.path
        # Casting AnyStr to str
        return str(dst_url.hostname) + path

    def get_complete_url(self, dst_url: str, src_url: str) -> str:
        """
        Get complete url with host, port, username, password if they are not provided in the `dst_url`
        """
        complete_url = urlparse(self.get_uri())
        _dst_url = urlparse(dst_url)
        _src_url = urlparse(src_url)

        path = self._get_url_path(dst_url=_dst_url, src_url=_src_url)

        final_url = complete_url._replace(path=path)

        return str(urlunparse(final_url))

    def write_using_smart_open(self, source_ref: DataStream | pd.DataFrame):
        """Write the source data from remote object i/o buffer to the dataset using smart open"""
        if isinstance(source_ref, DataStream):
            return self.write_from_file(source_ref=source_ref)
        elif isinstance(source_ref, pd.DataFrame):
            return self.write_from_dataframe(source_ref=source_ref)

    def write_from_file(self, source_ref: DataStream) -> str:
        """Write the remote object i/o buffer to the dataset using smart open
        :param source_ref: DataStream object of source dataset
        :return: File path that is the used for write pattern
        """
        mode = "wb" if self.read_as_binary(source_ref.actual_file.path) else "w"
        destination_file = self.dataset.path
        if self.dataset.is_pattern():
            destination_file = os.path.join(self.dataset.path, os.path.basename(source_ref.actual_filename))
        complete_url = self.get_complete_url(destination_file, source_ref.actual_file.path)
        with smart_open.open(complete_url, mode=mode, transport_params=self.transport_params) as stream:
            stream.write(source_ref.remote_obj_buffer.read())
        return destination_file

    def write_from_dataframe(self, source_ref: pd.DataFrame) -> str:
        """Write the dataframe to the SFTP dataset using smart open
        :param source_ref: DataStream object of source dataset
        :return: File path that is the used for write pattern
        """
        mode = "wb" if self.read_as_binary(self.dataset.path) else "w"
        complete_url = self.get_complete_url(self.dataset.path, "")
        with smart_open.open(complete_url, mode=mode, transport_params=self.transport_params) as stream:
            self.dataset.type.create_from_dataframe(stream=stream, df=source_ref)
        return self.dataset.path

    @property
    def openlineage_dataset_namespace(self) -> str:
        """
        Returns the open lineage dataset namespace as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        raise NotImplementedError

    @property
    def openlineage_dataset_name(self) -> str:
        """
        Returns the open lineage dataset name as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        raise NotImplementedError

    @property
    def size(self) -> int:
        """Return file size for SFTP location"""
        url = urlparse(self.dataset.path)
        conn = self.hook.get_conn()
        stat = conn.stat(url.netloc + url.path).st_size
        return int(stat) if stat else -1



from __future__ import annotations

import logging
from contextlib import contextmanager
from functools import cached_property
from tempfile import NamedTemporaryFile
from typing import Any, Iterator
from urllib.parse import urlparse, urlunparse

from airflow.providers.google.cloud.hooks.gcs import GCSHook, _parse_gcs_url

from universal_transfer_operator.constants import FileLocation, Location, TransferMode
from universal_transfer_operator.data_providers.filesystem.base import (
    BaseFilesystemProviders,
    Path,
    TempFile,
)
from universal_transfer_operator.datasets.file.base import File
from universal_transfer_operator.integrations.base import TransferIntegrationOptions


class GCSDataProvider(BaseFilesystemProviders):
    """
    DataProviders interactions with GS Dataset.
    """

    location_type = FileLocation.GS

    def __init__(
        self,
        dataset: File,
        transfer_params: TransferIntegrationOptions = TransferIntegrationOptions(),
        transfer_mode: TransferMode = TransferMode.NONNATIVE,
    ):
        super().__init__(
            dataset=dataset,
            transfer_params=transfer_params,
            transfer_mode=transfer_mode,
        )
        self.transfer_mapping = {
            Location.S3,
            Location.GS,
        }

    @cached_property
    def hook(self) -> GCSHook:
        """Return an instance of the database-specific Airflow hook."""
        return GCSHook(
            gcp_conn_id=self.dataset.conn_id,
            delegate_to=self.delegate_to,
            impersonation_chain=self.google_impersonation_chain,
        )

    def delete(self, path: str | None = None):
        """
        Delete a file/object if they exists
        """
        path = self.dataset.path if path is None else path
        url = urlparse(path)
        self.hook.delete(bucket_name=url.netloc, object_name=url.path.lstrip("/"))

    @property
    def transport_params(self) -> dict:
        """get GCS credentials for storage"""
        client = self.hook.get_conn()
        return {"client": client}

    @property
    def paths(self) -> list[str]:
        """Resolve GS file paths with prefix"""
        url = urlparse(self.dataset.path)
        prefix = url.path[1:]
        prefixes = self.hook.list(
            bucket_name=self.bucket_name,  # type: ignore
            prefix=prefix,
            delimiter=self.delimiter,
        )
        paths = [urlunparse((url.scheme, url.netloc, keys, "", "", "")) for keys in prefixes]
        return paths

    def check_if_exists(self, path: str | None = None) -> Any:
        """Return true if the dataset exists"""
        path = self.dataset.path if path is None else path
        url = urlparse(path)
        return self.hook.exists(bucket_name=url.netloc, object_name=url.path.lstrip("/"))

    @contextmanager
    def read_using_hook(self) -> Iterator[list[TempFile]]:
        """Read the file from dataset and write to local file location"""
        if not self.check_if_exists():
            raise ValueError(f"{self.dataset.path} doesn't exits")

        logging.info(
            "Getting list of the files. Bucket: %s; Delimiter: %s; Prefix: %s",
            self.bucket_name,  # type: ignore
            self.delimiter,
            self.prefix,
        )
        files = self.hook.list(
            bucket_name=self.bucket_name,  # type: ignore
            prefix=self.prefix,
            delimiter=self.delimiter,
        )

        try:
            local_file_paths = []
            if files:
                for file in files:
                    local_file_paths.append(self.download_file(file))
            yield local_file_paths
        finally:
            # Clean up the local files
            self.cleanup(local_file_paths)

    def write_using_hook(self, source_ref: list[TempFile]) -> list[str]:
        """Write the file from local file location to the dataset"""
        destination_objects = []
        if source_ref:
            for file in source_ref:
                destination_objects.append(self.upload_file(file))
            logging.info("All done, uploaded %d files to Google Cloud Storage", len(source_ref))
        else:
            logging.info("In sync, no files needed to be uploaded to Google Cloud Storage")
        return destination_objects

    def upload_file(self, file: TempFile):
        """Upload file to GCS and return path"""
        # There will always be a '/' before file because it is
        # enforced at instantiation time
        if file.tmp_file is None:
            raise ValueError("Required param `file.tmp_file` missing")
        dest_gcs_object = self.blob_name + file.actual_filename.name
        self.hook.upload(
            bucket_name=self.bucket_name,
            object_name=dest_gcs_object,
            filename=file.tmp_file.as_posix(),
            gzip=self.gzip,
        )
        return dest_gcs_object

    def download_file(self, file) -> TempFile:
        """Download file and save to temporary path."""
        _, _, file_name = file.rpartition("/")
        with NamedTemporaryFile(suffix=file_name, delete=False) as tmp_file:
            self.hook.download(
                bucket_name=self.bucket_name,
                object_name=file,
                filename=tmp_file.name,
            )
            return TempFile(tmp_file=Path(tmp_file.name), actual_filename=Path(file_name))

    @property
    def delegate_to(self) -> Any:
        return self.dataset.extra.get("delegate_to", None)

    @property
    def google_impersonation_chain(self) -> Any:
        return self.dataset.extra.get("google_impersonation_chain", None)

    @property
    def delimiter(self) -> Any:
        return self.dataset.extra.get("delimiter", None)

    @property
    def bucket_name(self) -> Any:
        bucket_name, _ = _parse_gcs_url(gsurl=self.dataset.path)
        return bucket_name

    @property
    def prefix(self) -> Any:
        return self.dataset.extra.get("prefix", None)

    @property
    def gzip(self) -> Any:
        return self.dataset.extra.get("gzip", False)

    @property
    def blob_name(self) -> Any:
        _, blob = _parse_gcs_url(gsurl=self.dataset.path)
        return blob

    @property
    def openlineage_dataset_namespace(self) -> str:
        """
        Returns the open lineage dataset namespace as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        raise NotImplementedError

    @property
    def openlineage_dataset_name(self) -> str:
        """
        Returns the open lineage dataset name as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        raise NotImplementedError

    @property
    def openlineage_dataset_uri(self) -> str:
        """
        Returns the open lineage dataset uri as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        raise NotImplementedError

    @property
    def size(self) -> int:
        """Return file size for GCS location"""
        url = urlparse(self.dataset.path)
        bucket_name = url.netloc
        object_name = url.path
        if object_name.startswith("/"):
            object_name = object_name[1:]
        return int(self.hook.get_size(bucket_name=bucket_name, object_name=object_name))



from __future__ import annotations

import importlib

from universal_transfer_operator.datasets.base import Dataset
from universal_transfer_operator.integrations.fivetran.destination.base import FivetranDestination
from universal_transfer_operator.utils import get_class_name

CUSTOM_FIVETRAN_DESTINATION_TYPE_TO_MODULE_PATH = {
    "snowflake": "universal_transfer_operator.integrations.fivetran.destination.snowflake"
}


def get_fivetran_destination(
    destination_dataset: Dataset,
    service: str,
    config: dict = {},
    destination_id: str | None = None,
    time_zone_offset: str | None = "-5",
    region: str | None = "US",
    run_setup_tests: bool | None = True,
) -> FivetranDestination:
    """
    Given a destination_dataset return the associated Fivetran Destination class.

    :param destination_dataset: Destination dataset
    :param service: services as per https://fivetran.com/docs/rest-api/destinations/config
    :param config: Configuration as per destination specified at https://fivetran.com/docs/rest-api/destinations/config
    :param destination_id: The unique identifier for the destination within the Fivetran system
    :param time_zone_offset: Determines the time zone for the Fivetran sync schedule.
    :param region: Data processing location. This is where Fivetran will operate and run computation on data.
    :param run_setup_tests: Specifies whether setup tests should be run automatically.
    """
    destination_conn_id = destination_dataset.conn_id  # type: ignore

    if destination_conn_id is None:
        raise ValueError("Connection id for destination_dataset is not specified.")

    from airflow.hooks.base import BaseHook

    destination_conn_type = BaseHook.get_connection(destination_conn_id).conn_type
    module_path = CUSTOM_FIVETRAN_DESTINATION_TYPE_TO_MODULE_PATH.get(destination_conn_type, None)
    if not module_path:
        raise KeyError(f"Connection type {destination_conn_type} not supported.")

    module = importlib.import_module(module_path)
    class_name = get_class_name(module_ref=module, suffix="Destination")
    fivetran_destination: FivetranDestination = getattr(module, class_name)(
        service=service,
        config=config,
        destination_id=destination_id,
        time_zone_offset=time_zone_offset,
        region=region,
        run_setup_tests=run_setup_tests,
    )
    return fivetran_destination



# TODO: Once the following PR is merged and released, we can
# remove the Fivetran hook: https://github.com/fivetran/airflow-provider-fivetran/pull/74
from __future__ import annotations

import json
from time import sleep
from typing import Any

import pendulum
import requests
from airflow.exceptions import AirflowException
from airflow.hooks.base import BaseHook
from requests import exceptions as requests_exceptions


class FivetranHook(BaseHook):
    """
    Fivetran API interaction hook.

    :param fivetran_conn_id: `Conn ID` of the Connection to be used to configure this hook.
    :param timeout_seconds: The amount of time in seconds the requests library will wait before timing out.
    :param retry_limit: The number of times to retry the connection in case of service outages.
    :param retry_delay: The number of seconds to wait between retries.
    """

    conn_name_attr = "fivetran_conn_id"
    default_conn_name = "fivetran_default"
    conn_type = "fivetran"
    hook_name = "Fivetran"
    api_user_agent = "airflow_provider_fivetran/1.1.4"
    api_protocol = "https"
    api_host = "api.fivetran.com"
    api_path_connectors = "v1/connectors/"
    api_metadata_path_connectors = "v1/metadata/connectors/"
    api_path_destinations = "v1/destinations/"
    api_path_groups = "v1/groups/"

    @staticmethod
    def get_ui_field_behaviour() -> dict:
        """Returns custom field behaviour"""
        return {
            "hidden_fields": ["schema", "port", "extra", "host"],
            "relabeling": {
                "login": "Fivetran API Key",
                "password": "Fivetran API Secret",
            },
            "placeholders": {
                "login": "api key",
                "password": "api secret",
            },
        }

    @staticmethod
    def _get_airflow_version() -> Any:
        """
        Fetch and return the current Airflow version from aws provider
        https://github.com/apache/airflow/blob/main/airflow/providers/amazon/aws/hooks/base_aws.py#L486
        """
        try:
            # This can be a circular import under specific configurations.
            # Importing locally to either avoid or catch it if it does happen.
            from airflow import __version__ as airflow_version

            return "-airflow_version/" + airflow_version
        except Exception:
            # Under no condition should an error here ever cause an issue for the user.
            return ""

    def __init__(
        self,
        fivetran_conn_id: str = "fivetran",
        fivetran_conn=None,
        timeout_seconds: int = 180,
        retry_limit: int = 3,
        retry_delay: float = 1.0,
    ) -> None:
        super().__init__(None)  # Passing None fixes a runtime problem in Airflow 1
        self.conn_id = fivetran_conn_id
        self.fivetran_conn = fivetran_conn
        self.timeout_seconds = timeout_seconds
        if retry_limit < 1:
            raise ValueError("Retry limit must be greater than equal to 1")
        self.retry_limit = retry_limit
        self.retry_delay = retry_delay

    def _do_api_call(self, endpoint_info, json=None):
        """
        Utility function to perform an API call with retries

        :param endpoint_info: Tuple of method and endpoint
        :param json: Parameters for this API call.
        """
        method, endpoint = endpoint_info
        if self.fivetran_conn is None:
            self.fivetran_conn = self.get_connection(self.conn_id)
        auth = (self.fivetran_conn.login, self.fivetran_conn.password)
        url = f"{self.api_protocol}://{self.api_host}/{endpoint}"

        headers = {"User-Agent": self.api_user_agent + self._get_airflow_version()}

        if method == "GET":
            request_func = requests.get
        elif method == "POST":
            request_func = requests.post
            headers.update({"Content-Type": "application/json;version=2"})
        elif method == "PATCH":
            request_func = requests.patch
            headers.update({"Content-Type": "application/json;version=2"})
        else:
            raise AirflowException("Unexpected HTTP Method: " + method)

        attempt_num = 1
        while True:
            try:
                response = request_func(
                    url,
                    data=json if method in ("POST", "PATCH") else None,
                    params=json if method in ("GET") else None,
                    auth=auth,
                    headers=headers,
                )
                response.raise_for_status()
                return response.json()
            except requests_exceptions.RequestException as e:
                if not _retryable_error(e):
                    # In this case, the user probably made a mistake.
                    # Don't retry.
                    raise AirflowException(
                        f"Response: {e.response.content}, Status Code: {e.response.status_code}"
                    )

                self._log_request_error(attempt_num, e)

            if attempt_num == self.retry_limit:
                raise AirflowException(f"API request to Fivetran failed {self.retry_limit} times. Giving up.")

            attempt_num += 1
            sleep(self.retry_delay)

    def _log_request_error(self, attempt_num: int, error: str) -> None:
        self.log.error(
            "Attempt %s API Request to Fivetran failed with reason: %s",
            attempt_num,
            error,
        )

    def _connector_ui_url(self, service_name, schema_name):
        return f"https://fivetran.com/dashboard/connectors/{service_name}/{schema_name}"

    def _connector_ui_url_logs(self, service_name, schema_name):
        return self._connector_ui_url(service_name, schema_name) + "/logs"

    def _connector_ui_url_setup(self, service_name, schema_name):
        return self._connector_ui_url(service_name, schema_name) + "/setup"

    def get_connector(self, connector_id) -> Any:
        """
        Fetches the detail of a connector.

        :param connector_id: Fivetran connector_id, found in connector settings page in the Fivetran user interface.
        """
        if connector_id == "":
            raise ValueError("No value specified for connector_id")
        endpoint = self.api_path_connectors + connector_id
        resp = self._do_api_call(("GET", endpoint))
        return resp["data"]

    def get_connector_schemas(self, connector_id) -> Any:
        """
        Fetches schema information of the connector.

        :param connector_id: Fivetran connector_id, found in connector settings page in the Fivetran user interface.
        """
        if connector_id == "":
            raise ValueError("No value specified for connector_id")
        endpoint = self.api_path_connectors + connector_id + "/schemas"
        resp = self._do_api_call(("GET", endpoint))
        return resp["data"]

    def get_metadata(self, connector_id, metadata) -> Any:
        """
        Fetches metadata for a given metadata string and connector. The Fivetran metadata API is currently in beta and
        available to all Fivetran users on the enterprise plan and above.

        :param connector_id: Fivetran connector_id, found in connector settings page in the Fivetran user interface.
        :param metadata: The string to return the type of metadata from the API
        """
        metadata_values = ("tables", "columns")
        if connector_id == "":
            raise ValueError("No value specified for connector_id")
        if metadata not in metadata_values:
            raise ValueError(f"Got {metadata} for param 'metadata', expected one of: {metadata_values}")
        endpoint = self.api_metadata_path_connectors + connector_id + "/" + metadata
        resp = self._do_api_call(("GET", endpoint))
        return resp["data"]

    def get_destinations(self, group_id) -> Any:
        """
        Fetches destination information for the given group.

        :param group_id: The Fivetran group ID, returned by a connector API call.
        """
        if group_id == "":
            raise ValueError("No value specified for group_id")
        endpoint = self.api_path_destinations + group_id
        resp = self._do_api_call(("GET", endpoint))
        return resp["data"]

    def get_groups(self, group_id) -> Any:
        """
        Fetches destination information for the given group.

        :param group_id: The Fivetran group ID, returned by a connector API call.
        """
        if group_id == "":
            raise ValueError("No value specified for connector_id")
        endpoint = self.api_path_groups + group_id
        resp = self._do_api_call(("GET", endpoint))
        return resp["data"]

    def check_connector(self, connector_id):
        """
        Ensures connector configuration has been completed successfully and is in a functional state.

        :param connector_id: Fivetran connector_id, found in connector settings page in the Fivetran user interface.
        """
        connector_details = self.get_connector(connector_id)
        service_name = connector_details["service"]
        schema_name = connector_details["schema"]
        setup_state = connector_details["status"]["setup_state"]
        if setup_state != "connected":
            raise AirflowException(
                f'Fivetran connector "{connector_id}" not correctly configured, '
                f"status: {setup_state}\nPlease see: "
                f"{self._connector_ui_url_setup(service_name, schema_name)}"
            )
        self.log.info(f"Connector type: {service_name}, connector schema: {schema_name}")
        self.log.info(f"Connectors logs at {self._connector_ui_url_logs(service_name, schema_name)}")
        return True

    def set_schedule_type(self, connector_id, schedule_type):
        """
        Set connector sync mode to switch sync control between API and UI.

        :param connector_id: Fivetran connector_id, found in connector settings page in the Fivetran user interface.
        :param schedule_type: Fivetran schedule type
        """
        endpoint = self.api_path_connectors + connector_id
        return self._do_api_call(("PATCH", endpoint), json.dumps({"schedule_type": schedule_type}))

    def prep_connector(self, connector_id, schedule_type):
        """
        Prepare the connector to run in Airflow by checking that it exists and is a good state, then update connector
        sync schedule type if changed.

        :param connector_id: Fivetran connector_id, found in connector settings page in the Fivetran user interface.
        :param schedule_type: Fivetran schedule type
        """
        self.check_connector(connector_id)
        if schedule_type not in {"manual", "auto"}:
            raise ValueError('schedule_type must be either "manual" or "auto"')
        if self.get_connector(connector_id)["schedule_type"] != schedule_type:
            return self.set_schedule_type(connector_id, schedule_type)
        return True

    def start_fivetran_sync(self, connector_id):
        connector_details = self.get_connector(connector_id)
        succeeded_at = connector_details["succeeded_at"]
        failed_at = connector_details["failed_at"]
        endpoint = self.api_path_connectors + connector_id
        if self._do_api_call(("GET", endpoint))["data"]["paused"] is True:
            self._do_api_call(("PATCH", endpoint), json.dumps({"paused": False}))
            if succeeded_at is None and failed_at is None:
                succeeded_at = str(pendulum.now())
        self._do_api_call(("POST", endpoint + "/force"))
        last_sync = (
            succeeded_at
            if self._parse_timestamp(succeeded_at) > self._parse_timestamp(failed_at)
            else failed_at
        )
        return last_sync

    def get_last_sync(self, connector_id, xcom=""):
        """
        Get the last time Fivetran connector completed a sync. Used with FivetranSensor to monitor sync completion.

        :param connector_id: Fivetran connector_id, found in connector settings page in the Fivetran user interface.
        :param xcom: Timestamp as string pull from FivetranOperator via XCOM
        """
        if xcom:
            last_sync = self._parse_timestamp(xcom)
        else:
            connector_details = self.get_connector(connector_id)
            succeeded_at = self._parse_timestamp(connector_details["succeeded_at"])
            failed_at = self._parse_timestamp(connector_details["failed_at"])
            last_sync = succeeded_at if succeeded_at > failed_at else failed_at
        return last_sync

    def get_sync_status(self, connector_id, previous_completed_at):
        """
        For sensor, return True if connector's 'succeeded_at' field has updated.

        :param connector_id: Fivetran connector_id, found in connector settings page in the Fivetran user interface.
        :param previous_completed_at: The last time the connector ran, collected on Sensor initialization.
        """
        # @todo Need logic here to tell if the sync is not running at all and not
        # likely to run in the near future.
        connector_details = self.get_connector(connector_id)
        succeeded_at = self._parse_timestamp(connector_details["succeeded_at"])
        failed_at = self._parse_timestamp(connector_details["failed_at"])
        current_completed_at = succeeded_at if succeeded_at > failed_at else failed_at

        # The only way to tell if a sync failed is to check if its latest
        # failed_at value is greater than then last known "sync completed at" value.
        if failed_at > previous_completed_at:
            service_name = connector_details["service"]
            schema_name = connector_details["schema"]
            raise AirflowException(
                f'Fivetran sync for connector "{connector_id}" failed; '
                f"please see logs at "
                f"{self._connector_ui_url_logs(service_name, schema_name)}"
            )

        sync_state = connector_details["status"]["sync_state"]
        self.log.info(f'Connector "{connector_id}": sync_state = {sync_state}')

        # Check if sync started by FivetranOperator has finished
        # indicated by new 'succeeded_at' timestamp
        if current_completed_at > previous_completed_at:
            self.log.info(f'Connector "{connector_id}": succeeded_at: {succeeded_at.to_iso8601_string()}')
            return True
        else:
            return False

    def _parse_timestamp(self, api_time):
        """
        Returns either the pendulum-parsed actual timestamp or a very out-of-date timestamp if not set

        :param api_time: timestamp format as returned by the Fivetran API.
        """
        return pendulum.parse(api_time) if api_time is not None else pendulum.from_timestamp(-1)

    def test_connection(self):
        """
        Ensures Airflow can reach Fivetran API
        """
        try:
            resp = self._do_api_call(("GET", "v1/users"))
            if resp["code"] == "Success":
                return True, "Fivetran connection test passed"
            else:
                return False, resp
        except Exception as e:
            return False, str(e)


def _retryable_error(exception) -> bool:
    return (
        isinstance(
            exception,
            (requests_exceptions.ConnectionError, requests_exceptions.Timeout),
        )
        or exception.response is not None
        and exception.response.status_code >= 500
    )



from __future__ import annotations

import io

import pandas as pd

from universal_transfer_operator.constants import FileType as FileTypeConstants
from universal_transfer_operator.datasets.file.types.base import FileTypes


class CSVFileTypes(FileTypes):
    """Concrete implementation to handle CSV file type"""

    # We need skipcq because it's a method overloading so we don't want to make it a static method
    def export_to_dataframe(
        self, stream, columns_names_capitalization="original", **kwargs
    ) -> pd.DataFrame:  # skipcq PYL-R0201
        """
        Read csv file from one of the supported locations and return dataframe.

        :param stream: file stream object
        :param columns_names_capitalization: determines whether to convert all columns to lowercase/uppercase
            in the resulting dataframe
        """
        return pd.read_csv(stream, **kwargs)

    # We need skipcq because it's a method overloading so we don't want to make it a static method
    def create_from_dataframe(self, df: pd.DataFrame, stream: io.TextIOWrapper) -> None:  # skipcq PYL-R0201
        """
        Write csv file to one of the supported locations

        :param df: pandas dataframe
        :param stream: file stream object
        """
        df.to_csv(stream, index=False)

    @property
    def name(self):
        return FileTypeConstants.CSV



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any

import attr
from airflow.hooks.base import BaseHook

from universal_transfer_operator.datasets.base import Dataset
from universal_transfer_operator.utils import TransferParameters


@attr.define
class TransferIntegrationOptions(TransferParameters):
    """TransferIntegrationOptions for transfer integration configuration"""

    conn_id: str = attr.field(default="")


class TransferIntegration(ABC):
    """Basic implementation of a third party transfer."""

    def __init__(
        self,
        transfer_params: TransferIntegrationOptions = attr.field(
            factory=TransferIntegrationOptions,
            converter=lambda val: TransferIntegrationOptions(**val) if isinstance(val, dict) else val,
        ),
    ):
        self.transfer_params = transfer_params
        # transfer mapping creates a mapping between various sources and destination, where
        # transfer is possible using the integration
        self.transfer_mapping: dict[str, str] = {}
        # TODO: add method for validation, transfer mapping, transfer params etc

    @property
    def hook(self) -> BaseHook:
        """Return an instance of the database-specific Airflow hook."""
        raise NotImplementedError

    @abstractmethod
    def transfer_job(self, source_dataset: Dataset, destination_dataset: Dataset) -> Any:
        """
        Loads data from source dataset to the destination using ingestion config
        """
        raise NotImplementedError



from __future__ import annotations

import io
import os
from abc import abstractmethod
from pathlib import Path
from typing import Iterator

import attr
import pandas as pd
import smart_open
from airflow.hooks.base import BaseHook

from universal_transfer_operator.constants import FileType, Location, TransferMode
from universal_transfer_operator.data_providers.base import DataProviders, DataStream
from universal_transfer_operator.datasets.file.base import File
from universal_transfer_operator.datasets.file.types import create_file_type
from universal_transfer_operator.universal_transfer_operator import TransferIntegrationOptions
from universal_transfer_operator.utils import get_dataset_connection_type


@attr.define
class TempFile:
    tmp_file: Path | None
    actual_filename: Path


class BaseFilesystemProviders(DataProviders[File]):
    """BaseFilesystemProviders represent all the DataProviders interactions with File system."""

    def __init__(
        self,
        dataset: File,
        transfer_mode,
        transfer_params: TransferIntegrationOptions = TransferIntegrationOptions(),
    ):
        self.dataset = dataset
        self.transfer_params = transfer_params
        self.transfer_mode = transfer_mode
        self.transfer_mapping: set[Location] = set()
        self.LOAD_DATA_NATIVELY_FROM_SOURCE: dict = {}
        super().__init__(dataset=self.dataset)

    def __repr__(self):
        return f'{self.__class__.__name__}(conn_id="{self.dataset.conn_id})'

    @property
    def hook(self) -> BaseHook:
        """Return an instance of the database-specific Airflow hook."""
        raise NotImplementedError

    @property
    @abstractmethod
    def paths(self) -> list[str]:
        """Resolve patterns in path"""
        raise NotImplementedError

    def delete(self, path: str):  # type: ignore
        """
        Delete a file/object if they exists
        """
        raise NotImplementedError

    @property
    def transport_params(self) -> dict | None:  # skipcq: PYL-R0201
        """Get credentials required by smart open to access files"""
        return None

    def check_if_exists(self) -> bool:
        """Return true if the dataset exists"""
        raise NotImplementedError

    def exists(self) -> bool:
        return self.check_if_exists()

    def check_if_transfer_supported(self, source_dataset: File) -> bool:
        """
        Checks if the transfer is supported from source to destination based on source_dataset.
        """
        source_connection_type = get_dataset_connection_type(source_dataset)
        return Location(source_connection_type) in self.transfer_mapping

    def read(self) -> Iterator[DataStream]:
        """Read the remote or local file dataset and returns i/o buffers"""
        if self.transfer_mode == TransferMode.NATIVE:
            return iter(
                [
                    DataStream(
                        actual_file=self.dataset, remote_obj_buffer=io.BytesIO(), actual_filename=Path("")
                    )
                ]
            )
        else:
            return self.read_using_smart_open()

    def read_using_smart_open(self) -> Iterator[DataStream]:
        """Read the file dataset using smart open returns i/o buffer"""
        files = self.paths
        for file in files:
            yield DataStream(
                remote_obj_buffer=self._convert_remote_file_to_byte_stream(file),
                actual_filename=Path(file),
                actual_file=self.dataset,
            )

    def _convert_remote_file_to_byte_stream(self, file: str) -> io.IOBase:
        """
        Read file from all supported location and convert them into a buffer that can be streamed into other data
        structures.

        :returns: an io object that can be streamed into a dataframe (or other object)
        """
        mode = "rb" if self.read_as_binary(file) else "r"
        remote_obj_buffer = io.BytesIO() if self.read_as_binary(file) else io.StringIO()
        with smart_open.open(file, mode=mode, transport_params=self.transport_params) as stream:
            remote_obj_buffer.write(stream.read())
            remote_obj_buffer.seek(0)
            return remote_obj_buffer

    def write(self, source_ref: DataStream | pd.DataFrame) -> str:
        """
        Write the data from local reference location or a dataframe to the filesystem dataset or database dataset

        :param source_ref: Source DataStream object which will be used to read data
        """
        return self.write_using_smart_open(source_ref=source_ref)

    def write_using_smart_open(self, source_ref: DataStream | pd.DataFrame) -> str:
        """Write the source data from remote object i/o buffer to the dataset using smart open"""
        if isinstance(source_ref, DataStream):
            # `source_ref` can be a dataframe for all the filetypes we can create a dataframe for like -
            # CSV, JSON, NDJSON, and Parquet or SQL Tables. This gives us the option to perform various
            # functions on the data on the fly, like filtering or changing the file format altogether. For other
            # files whose content cannot be converted to dataframe like - zip or image, we get a DataStream object.
            return self.write_from_file(source_ref)
        else:
            return self.write_from_dataframe(source_ref)

    def write_from_file(self, source_ref: DataStream) -> str:
        """Write the remote object i/o buffer to the dataset using smart open
        :param source_ref: DataStream object of source dataset
        :return: File path that is the used for write pattern
        """
        mode = "wb" if self.read_as_binary(source_ref.actual_file.path) else "w"

        destination_file = self.dataset.path
        # check if destination dataset is folder or file pattern
        if self.dataset.is_pattern():
            destination_file = os.path.join(self.dataset.path, os.path.basename(source_ref.actual_filename))

        with smart_open.open(destination_file, mode=mode, transport_params=self.transport_params) as stream:
            stream.write(source_ref.remote_obj_buffer.read())
        return destination_file

    def write_from_dataframe(self, source_ref: pd.DataFrame) -> str:
        """Write the dataframe to the SFTP dataset using smart open
        :param source_ref: DataStream object of source dataset
        :return: File path that is the used for write pattern
        """
        mode = "wb" if self.read_as_binary(self.dataset.path) else "w"

        destination_file = self.dataset.path
        with smart_open.open(destination_file, mode=mode, transport_params=self.transport_params) as stream:
            self.dataset.type.create_from_dataframe(stream=stream, df=source_ref)
        return destination_file

    def read_as_binary(self, file: str) -> bool:
        """
        Checks if file has to be read as binary or as string i/o.

        :return: True or False
        """
        try:
            filetype = create_file_type(
                path=file,
                filetype=self.dataset.filetype,
                normalize_config=self.dataset.normalize_config,
            )
        except ValueError:
            # return True when the extension of file is not supported
            # Such file can be read as binary and transferred.
            return True

        read_as_non_binary = {FileType.CSV, FileType.JSON, FileType.NDJSON}
        if filetype in read_as_non_binary:
            return False
        return True

    @staticmethod
    def cleanup(file_list: list[TempFile]) -> None:
        """Cleans up the temporary files created"""
        for file in file_list:
            if os.path.exists(file.actual_filename):
                os.remove(file.actual_filename)

    @property
    def openlineage_dataset_namespace(self) -> str:
        """
        Returns the open lineage dataset namespace as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        raise NotImplementedError

    @property
    def openlineage_dataset_name(self) -> str:
        """
        Returns the open lineage dataset name as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        raise NotImplementedError

    @property
    def openlineage_dataset_uri(self) -> str:
        """
        Returns the open lineage dataset uri as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def size(self) -> int:
        """Return the size in bytes of the given file"""
        raise NotImplementedError

    @property
    def snowflake_stage_path(self) -> str:
        """
        Get the altered path if needed for stage creation in snowflake stage creation
        """
        return self.dataset.path

    def is_native_path_available(
        self,
        source_dataset: File,  # skipcq PYL-W0613, PYL-R0201
    ) -> bool:
        """
        Check if there is an optimised path for source to destination.

        :param source_dataset: File from which we need to transfer data
        """
        return False



from __future__ import annotations

from typing import Any

from attr import define, field

from universal_transfer_operator.datasets.base import Dataset


@define
class Dataframe(Dataset):
    """
    Repersents all dataframe dataset.
    Intended to be used within library.
    :param name: name of dataframe

    """

    name: str = field(default="")
    dataframe: Any = field(default=None)
    uri: str = field(init=None)  # type: ignore

    # TODO: define the name and namespace for dataframe

    @uri.default
    def _path_to_dataset_uri(self) -> str:
        """Build a URI to be passed to Dataset obj introduced in Airflow 2.4"""
        return self.name



from __future__ import annotations

from typing import Iterator

import pandas as pd
from universal_transfer_operator.data_providers.base import DataProviders, DataStream
from universal_transfer_operator.datasets.dataframe.base import Dataframe


class DataframeProvider(DataProviders[Dataframe]):
    """Base class to import dataframe implementation. We intend to support different implementation of dataframes."""

    def __init__(
        self,
        dataset: Dataframe,
    ) -> None:
        self.dataset = dataset
        super().__init__(dataset=self.dataset)

    def read(self) -> Iterator[Dataframe]:
        """Read from dataframe dataset and write to local reference locations or dataframes"""
        raise NotImplementedError

    def write(self, source_ref: pd.DataFrame | DataStream) -> DataframeProvider:  # type: ignore
        """Write the data to the dataframe dataset or filesystem dataset"""
        raise NotImplementedError

    def serialize(self):
        """Store in the metadata DB if Dataframe"""
        raise NotImplementedError

    @staticmethod
    def deserialize(data: dict, version: int):
        """Extract from metadata DB"""
        raise NotImplementedError

    def is_native_path_available(
        self,
        source_dataset: Dataframe,  # skipcq PYL-W0613, PYL-R0201
    ) -> bool:
        """
        Check if there is an optimised path for source to destination.

        :param source_dataset: Dataframe from which we need to transfer data
        """
        return False



from __future__ import annotations

import importlib
from typing import Type, cast

from airflow.hooks.base import BaseHook

from universal_transfer_operator.constants import IngestorSupported
from universal_transfer_operator.integrations.base import TransferIntegration, TransferIntegrationOptions
from universal_transfer_operator.utils import get_class_name

CUSTOM_INGESTION_TYPE_TO_MODULE_PATH = {
    "fivetran": "universal_transfer_operator.integrations.fivetran.fivetran"
}


def get_transfer_integration(transfer_params: TransferIntegrationOptions) -> TransferIntegration:
    """
    Given a transfer_params return the associated TransferIntegrations class instance.

    :param transfer_params: kwargs to be used by methods involved in transfer using FiveTran.
    """
    conn_id = get_conn_id(transfer_params=transfer_params)
    return get_ingestor_class(conn_id=conn_id)(transfer_params)


def get_ingestor_class(conn_id: str) -> type[TransferIntegration]:
    """
    Given a transfer_params return the associated TransferIntegrations class.

    :param transfer_params: kwargs to be used by methods involved in transfer using FiveTran.
    """
    thirdparty_conn_type = BaseHook.get_connection(conn_id).conn_type
    if thirdparty_conn_type not in {item.value for item in IngestorSupported}:
        raise ValueError("Ingestion platform not yet supported.")

    module_path = CUSTOM_INGESTION_TYPE_TO_MODULE_PATH[thirdparty_conn_type]
    module = importlib.import_module(module_path)
    class_name = get_class_name(module_ref=module, suffix="Integration")
    return cast(Type[TransferIntegration], getattr(module, class_name))


def get_ingestor_option_class(conn_id: str) -> type[TransferIntegrationOptions]:
    """
    Get options class for fivetran
    """
    class_ref = get_ingestor_class(conn_id)
    return getattr(class_ref, "OPTIONS_CLASS", TransferIntegrationOptions)


def get_conn_id(transfer_params: TransferIntegrationOptions | dict) -> str:
    """
    Get conn_id from transfer_params
    """
    try:
        conn_id = transfer_params.conn_id  # type: ignore
    except AttributeError:
        conn_id = transfer_params.get("conn_id")  # type: ignore
    if conn_id is None:
        raise ValueError("Connection id for integration is not specified.")
    return conn_id



from __future__ import annotations

import json
import logging
import uuid
from typing import Any

import boto3
from airflow.hooks.base import BaseHook
from airflow.providers.amazon.aws.hooks.s3 import S3Hook

from universal_transfer_operator.integrations.fivetran.connector.base import Dataset, FivetranConnector
from universal_transfer_operator.settings import FIVETRAN_AWS_VPC_ACCOUNT_ID


class S3Connector(FivetranConnector):
    """
    Fivetran connector details. More details at: https://fivetran.com/docs/rest-api/connectors/config#amazons3

    :param connector_id: The unique identifier for the connector within the Fivetran system
    :param service: services as per https://fivetran.com/docs/rest-api/connectors#createaconnector
    :param config: Configuration as per destination specified at https://fivetran.com/docs/rest-api/connectors/config
    :param paused: Specifies whether the connector is paused. Defaults to False.
    :param pause_after_trial: Specifies whether the connector should be paused after the free trial period has ended.
     Defaults to False.
    :param sync_frequency: The connector sync frequency in minutes Enum: "5" "15" "30" "60" "120" "180" "360" "480"
     "720" "1440".
    :param daily_sync_time: Defines the sync start time when the sync frequency is already set
    :param schedule_type: Define the schedule type
    :param connect_card_config: Connector card configuration
    :param trust_certificates: Specifies whether we should trust the certificate automatically. The default value is
     FALSE. If a certificate is not trusted automatically,
    :param trust_fingerprints: Specifies whether we should trust the SSH fingerprint automatically. The default value
     is FALSE.
    :param run_setup_tests: Specifies whether the setup tests should be run automatically. The default value is TRUE.
    """

    def __init__(
        self,
        connector_id: str | None,
        service: str,
        config: dict,
        paused: bool = False,
        pause_after_trial: bool = False,
        sync_frequency: str | None = None,
        daily_sync_time: str | None = None,
        schedule_type: str | None = None,
        trust_certificates: bool = False,
        trust_fingerprints: bool = False,
        run_setup_tests: bool = True,
        connect_card_config: dict | None = None,
    ):
        self.connector_id = connector_id
        self.service = service
        self.config = config
        self.paused = paused
        self.pause_after_trial = pause_after_trial
        self.sync_frequency = sync_frequency
        self.daily_sync_time = daily_sync_time
        self.schedule_type = schedule_type
        self.trust_certificates = trust_certificates
        self.trust_fingerprints = trust_fingerprints
        self.run_setup_tests = run_setup_tests
        self.connect_card_config = connect_card_config
        super().__init__(
            connector_id=self.connector_id,
            service=self.service,
            config=self.config,
            paused=self.paused,
            pause_after_trial=self.pause_after_trial,
            sync_frequency=self.sync_frequency,
            daily_sync_time=self.daily_sync_time,
            schedule_type=self.schedule_type,
            trust_certificates=self.trust_certificates,
            trust_fingerprints=self.trust_fingerprints,
            run_setup_tests=self.run_setup_tests,
            connect_card_config=self.connect_card_config,
        )

    def to_dict(self) -> dict:
        """
        Convert options class to dict
        """
        return vars(self)

    def _hook(self, conn_id: str, verify: Any, transfer_config_args: Any, s3_extra_args: Any) -> S3Hook:
        """Return an instance of the database-specific Airflow hook."""
        return S3Hook(
            aws_conn_id=conn_id,
            verify=verify,
            transfer_config_args=transfer_config_args,
            extra_args=s3_extra_args,
        )

    def map_airflow_connection_to_fivetran(
        self,
        source_dataset: Dataset,
        destination_dataset: Dataset,
        group_id: str,
    ) -> dict[str, str]:
        """
        Maps the airflow connection details to fivetran config for destination

        :param source_dataset: Source dataset
        :param destination_dataset: Destination dataset
        :param group_id: Group id in fivetran system
        """
        conn = BaseHook().get_connection(source_dataset.conn_id)  # type: ignore

        access_key = conn.login
        secret_key = conn.password

        if not access_key:
            raise ConnectionError(f"AWS_ACCESS_KEY_ID is not configured for {source_dataset.conn_id}")  # type: ignore

        if not secret_key:
            raise ConnectionError(f"AWS_SECRET_ACCESS_KEY is not configured {source_dataset.conn_id}")  # type: ignore

        s3_hook = self._hook(
            conn_id=source_dataset.conn_id,  # type: ignore
            verify=self._verify(source_dataset),
            transfer_config_args=self._transfer_config_args(source_dataset),
            s3_extra_args=self._s3_extra_args(source_dataset),
        )

        bucket_name = self._bucket_name(s3_hook=s3_hook, dataset=source_dataset)
        # Create IAM policy. Read more at: Read more at: https://fivetran.com/docs/files/amazon-s3/setup-guide
        iam_policy = self._create_aws_policy(
            bucket_name=bucket_name, access_key=access_key, secret_key=secret_key
        )

        # Create IAM Role. Read more at: Read more at: https://fivetran.com/docs/files/amazon-s3/setup-guide
        iam_role = self._create_role(
            role_name=self._generate_unique_fivetran_name(),
            fivetran_aws_vpc_account_id=str(FIVETRAN_AWS_VPC_ACCOUNT_ID),
            group_id=group_id,
            access_key=access_key,
            secret_key=secret_key,
        )

        # Attach IAM policy to role. Read more at: Read more at: https://fivetran.com/docs/files/amazon-s3/setup-guide
        self._attach_policy_to_role(
            role_name=iam_role["Role"]["RoleName"],
            policy_arn=iam_policy["Policy"]["Arn"],
            access_key=access_key,
            secret_key=secret_key,
        )

        connection_details = {
            "schema": destination_dataset.metadata.schema,  # type: ignore
            "table": destination_dataset.name,  # type: ignore
            "external_id": group_id,
            "is_public": self.config.get("is_public"),
            "role_arn": iam_role["Role"]["Arn"],
            "bucket": bucket_name,
            "prefix": self._prefix(dataset=source_dataset),
            "file_type": self.config.get("file_type"),
            "compression": self.config.get("compression"),
            "on_error": self.config.get("on_error"),
        }

        return connection_details

    @staticmethod
    def _verify(dataset: Dataset) -> Any:
        return dataset.extra.get("verify")  # type: ignore

    @staticmethod
    def _transfer_config_args(dataset: Dataset) -> Any:
        return dataset.extra.get("transfer_config_args", {})  # type: ignore

    @staticmethod
    def _s3_extra_args(dataset: Dataset) -> Any:
        return dataset.extra.get("s3_extra_args", {})  # type: ignore

    @staticmethod
    def _bucket_name(s3_hook: S3Hook, dataset: Dataset) -> Any:
        bucket_name, _ = s3_hook.parse_s3_url(dataset.path)  # type: ignore
        return bucket_name

    @staticmethod
    def _s3_key(s3_hook: S3Hook, dataset: Dataset) -> Any:
        _, key = s3_hook.parse_s3_url(dataset.path)  # type: ignore
        return key

    @staticmethod
    def _prefix(dataset: Dataset) -> Any:
        return dataset.extra.get("prefix", None)  # type: ignore

    @staticmethod
    def _delimiter(dataset: Dataset) -> Any:
        return dataset.extra.get("delimiter", None)  # type: ignore

    @staticmethod
    def _generate_unique_fivetran_name() -> str:
        """Generates the unique fivetran name based on uuid."""
        uuid_value = str(uuid.uuid4())[:8]
        unique_policy_name = f"fivetran-{uuid_value}"
        return unique_policy_name

    def _create_aws_policy(self, bucket_name: str, access_key: str, secret_key: str):
        """
        Creates the AWS IAM policy to read and list the S3 Bucket.
        Read more at: https://fivetran.com/docs/files/amazon-s3/setup-guide

        :param bucket_name: Name of bucket
        :param access_key: AWS access key
        :param secret_key: AWS secret key
        """
        # Create IAM client
        iam = boto3.client("iam", aws_access_key_id=access_key, aws_secret_access_key=secret_key)

        # Create a policy
        my_managed_policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Action": ["s3:GetObject", "s3:ListBucket"],
                    "Resource": [f"arn:aws:s3:::{bucket_name}/*", f"arn:aws:s3:::{bucket_name}"],
                }
            ],
        }
        from botocore.exceptions import ClientError

        try:
            response = iam.create_policy(
                PolicyName=self._generate_unique_fivetran_name(), PolicyDocument=json.dumps(my_managed_policy)
            )
        except ClientError:
            logging.error(f"Couldn't create policy with config {my_managed_policy}")
            raise
        return response

    @staticmethod
    def _create_role(
        role_name: str, fivetran_aws_vpc_account_id: str, group_id: str, access_key: str, secret_key: str
    ):
        """
        Creates a role that lets a list of specified services assume the role.
        Read more at: https://fivetran.com/docs/files/amazon-s3/setup-guide

        :param role_name: The name of the role.
        :param fivetran_aws_vpc_account_id: Fivetran’s AWS VPC Account ID
        :param group_id: Group id in fivetran system
        :param access_key: AWS access key
        :param secret_key: AWS secret key
        """
        # Create IAM client
        iam = boto3.client("iam", aws_access_key_id=access_key, aws_secret_access_key=secret_key)

        trust_policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Principal": {"AWS": f"arn:aws:iam::{fivetran_aws_vpc_account_id}:root"},
                    "Action": "sts:AssumeRole",
                    "Condition": {"StringEquals": {"sts:ExternalId": f"{group_id}"}},
                }
            ],
        }
        from botocore.exceptions import ClientError

        try:
            role = iam.create_role(RoleName=role_name, AssumeRolePolicyDocument=json.dumps(trust_policy))
            logging.info(f"role created with name {role_name}")
        except ClientError:
            logging.error(f"Couldn't create role {role_name}")
            raise
        else:
            return role

    @staticmethod
    def _attach_policy_to_role(role_name: str, policy_arn: str, access_key: str, secret_key: str):
        """
        Attaches a policy to a role.

        :param role_name: The name of the role. **Note** this is the name, not the ARN.
        :param policy_arn: The ARN of the policy.
        :param access_key: AWS access key
        :param secret_key: AWS secret key
        """
        iam = boto3.client("iam", aws_access_key_id=access_key, aws_secret_access_key=secret_key)
        from botocore.exceptions import ClientError

        try:
            # Attach a role policy
            iam.attach_role_policy(PolicyArn=policy_arn, RoleName=role_name)
        except ClientError:
            logging.error(f"Couldn't attach policy {policy_arn} to role {role_name}")
            raise



from __future__ import annotations

from abc import abstractmethod
from typing import Any

from universal_transfer_operator.integrations.fivetran.connector import Dataset

airflow_connection_type_to_fivetran_connector_service_mapping = {"s3": "s3", "aws": "s3"}


class FivetranConnector:
    """
    Fivetran connector details. More details at: https://fivetran.com/docs/rest-api/connectors#createaconnector

    :param connector_id: The unique identifier for the connector within the Fivetran system
    :param service: services as per https://fivetran.com/docs/rest-api/connectors#createaconnector
    :param config: Configuration as per destination specified at https://fivetran.com/docs/rest-api/connectors/config
    :param paused: Specifies whether the connector is paused. Defaults to False.
    :param pause_after_trial: Specifies whether the connector should be paused after the free trial period has ended.
     Defaults to False.
    :param sync_frequency: The connector sync frequency in minutes Enum: "5" "15" "30" "60" "120" "180" "360" "480"
     "720" "1440".
    :param daily_sync_time: Defines the sync start time when the sync frequency is already set
    :param schedule_type: Define the schedule type
    :param connect_card_config: Connector card configuration
    :param trust_certificates: Specifies whether we should trust the certificate automatically. The default value is
     FALSE. If a certificate is not trusted automatically,
    :param trust_fingerprints: Specifies whether we should trust the SSH fingerprint automatically. The default value
     is FALSE.
    :param run_setup_tests: Specifies whether the setup tests should be run automatically. The default value is TRUE.
    """

    def __init__(
        self,
        config: dict = {},
        paused: bool = False,
        pause_after_trial: bool = False,
        sync_frequency: str | None = None,
        daily_sync_time: str | None = None,
        schedule_type: str | None = None,
        trust_certificates: bool = False,
        trust_fingerprints: bool = False,
        run_setup_tests: bool = True,
        connect_card_config: dict | None = None,
        connector_id: str | None = None,
        service: str = "",
    ):
        self.connector_id = connector_id
        self.service = service
        self.config = config
        self.paused = paused
        self.pause_after_trial = pause_after_trial
        self.sync_frequency = sync_frequency
        self.daily_sync_time = daily_sync_time
        self.schedule_type = schedule_type
        self.trust_certificates = trust_certificates
        self.trust_fingerprints = trust_fingerprints
        self.run_setup_tests = run_setup_tests
        self.connect_card_config = connect_card_config

    def to_dict(self) -> dict:
        """
        Convert options class to dict
        """
        return vars(self)

    def create_config(
        self,
        source_dataset: Dataset,
        destination_dataset: Dataset,
        group_id: str,
    ) -> Any:
        """
        Creates config based on connection details and config

        :param source_dataset: Source dataset
        :param destination_dataset: Destination dataset
        :param group_id: Group id in fivetran system
        """
        mapped_connection_details = self.map_airflow_connection_to_fivetran(
            source_dataset=source_dataset, destination_dataset=destination_dataset, group_id=group_id
        )
        return self.config | mapped_connection_details  # type: ignore

    @abstractmethod
    def map_airflow_connection_to_fivetran(
        self, source_dataset: Dataset, destination_dataset: Dataset, group_id: str
    ) -> dict[str, str]:
        """
        Maps the airflow connection details to fivetran config for source connector

        :param source_dataset: Source dataset
        :param destination_dataset: Destination dataset
        :param group_id: Group id in fivetran system
        """
        raise NotImplementedError



from __future__ import annotations

import importlib

from universal_transfer_operator.datasets.base import Dataset
from universal_transfer_operator.integrations.fivetran.connector.base import FivetranConnector
from universal_transfer_operator.utils import get_class_name

CUSTOM_FIVETRAN_CONNECTOR_TYPE_TO_MODULE_PATH = {
    "s3": "universal_transfer_operator.integrations.fivetran.connector.aws.s3",
    "aws": "universal_transfer_operator.integrations.fivetran.connector.aws.s3",
}


def get_fivetran_connector(
    source_dataset: Dataset,
    connector_id: str | None,
    service: str,
    config: dict = {},
    paused: bool = False,
    pause_after_trial: bool = False,
    sync_frequency: str | None = None,
    daily_sync_time: str | None = None,
    schedule_type: str | None = None,
    trust_certificates: bool = False,
    trust_fingerprints: bool = False,
    run_setup_tests: bool = True,
    connect_card_config: dict | None = None,
) -> FivetranConnector:
    """
    Given a source_dataset return the associated Fivetran Connector class.

    :param connector_id: The unique identifier for the connector within the Fivetran system
    :param service: services as per https://fivetran.com/docs/rest-api/connectors#createaconnector
    :param config: Configuration as per destination specified at https://fivetran.com/docs/rest-api/connectors/config
    :param paused: Specifies whether the connector is paused. Defaults to False.
    :param pause_after_trial: Specifies whether the connector should be paused after the free trial period has ended.
     Defaults to False.
    :param sync_frequency: The connector sync frequency in minutes Enum: "5" "15" "30" "60" "120" "180" "360" "480"
     "720" "1440".
    :param daily_sync_time: Defines the sync start time when the sync frequency is already set
    :param schedule_type: Define the schedule type
    :param connect_card_config: Connector card configuration
    :param trust_certificates: Specifies whether we should trust the certificate automatically. The default value is
     FALSE. If a certificate is not trusted automatically,
    :param trust_fingerprints: Specifies whether we should trust the SSH fingerprint automatically. The default value
     is FALSE.
    :param run_setup_tests: Specifies whether the setup tests should be run automatically. The default value is TRUE.
    """
    source_conn_id = source_dataset.conn_id  # type: ignore

    if source_conn_id is None:
        raise ValueError("Connection id for source_dataset is not specified.")

    from airflow.hooks.base import BaseHook

    source_conn_type = BaseHook.get_connection(source_conn_id).conn_type
    module_path = CUSTOM_FIVETRAN_CONNECTOR_TYPE_TO_MODULE_PATH.get(source_conn_type, None)
    if not module_path:
        raise KeyError(f"Connection type {source_conn_type} not supported.")

    module = importlib.import_module(module_path)
    class_name = get_class_name(module_ref=module, suffix="Connector")
    fivetran_connector: FivetranConnector = getattr(module, class_name)(
        connector_id=connector_id,
        service=service,
        config=config,
        paused=paused,
        pause_after_trial=pause_after_trial,
        sync_frequency=sync_frequency,
        daily_sync_time=daily_sync_time,
        schedule_type=schedule_type,
        trust_certificates=trust_certificates,
        trust_fingerprints=trust_fingerprints,
        run_setup_tests=run_setup_tests,
        connect_card_config=connect_card_config,
    )
    return fivetran_connector



from __future__ import annotations

import pathlib

from universal_transfer_operator.constants import FileType as FileTypeConstants
from universal_transfer_operator.datasets.file.types.base import FileTypes
from universal_transfer_operator.datasets.file.types.csv import CSVFileTypes
from universal_transfer_operator.datasets.file.types.json import JSONFileTypes
from universal_transfer_operator.datasets.file.types.ndjson import NDJsonFileTypes
from universal_transfer_operator.datasets.file.types.parquet import ParquetFileTypes


def create_file_type(
    path: str,
    filetype: FileTypeConstants | None = None,
    normalize_config: dict | None = None,
) -> FileTypes:
    """Factory method to create FileType super objects based on the file extension in path or filetype specified."""
    filetype_to_class: dict[FileTypeConstants, type[FileTypes]] = {
        FileTypeConstants.CSV: CSVFileTypes,
        FileTypeConstants.JSON: JSONFileTypes,
        FileTypeConstants.NDJSON: NDJsonFileTypes,
        FileTypeConstants.PARQUET: ParquetFileTypes,
    }
    if not filetype:
        filetype = get_filetype(path)

    try:
        return filetype_to_class[filetype](path=path, normalize_config=normalize_config)
    except KeyError:
        raise ValueError(
            f"Non supported file type provided {filetype}, file_type should be among {', '.join(FileTypeConstants)}."
        )


def get_filetype(filepath: str | pathlib.PosixPath) -> FileTypeConstants:
    """
    Return a FileType given the filepath. Uses a native strategy, using the file extension.
    :param filepath: URI or Path to a file
    :type filepath: str or pathlib.PosixPath
    :return: The filetype (e.g. csv, ndjson, json, parquet)
    :rtype: universal_transfer_operator.constants.FileType
    """
    if isinstance(filepath, pathlib.PosixPath):
        extension = filepath.suffix[1:]
    else:
        extension = ""
        tokenized_path = filepath.split(".")
        if len(tokenized_path) > 1:
            extension = tokenized_path[-1]

    if extension == "":
        raise ValueError(
            f"Missing file extension, cannot automatically determine filetype from path '{filepath}'."
            f" Please pass the 'filetype' param with the explicit filetype (e.g. csv, ndjson, etc.)."
        )

    try:
        return FileTypeConstants(extension)
    except ValueError:
        raise ValueError(f"Unsupported filetype '{extension}' from file '{filepath}'.")



from __future__ import annotations

import logging
import os
import random
import string
from dataclasses import dataclass, field
from importlib.metadata import version
from typing import Sequence
from urllib.parse import urlparse

import attr
import pandas as pd
from airflow.providers.snowflake.hooks.snowflake import SnowflakeHook
from packaging import version as packaging_version
from snowflake.connector import pandas_tools
from snowflake.connector.errors import (
    ProgrammingError,
)

from universal_transfer_operator.constants import (
    DEFAULT_CHUNK_SIZE,
    FileLocation,
    FileType,
    LoadExistStrategy,
)
from universal_transfer_operator.data_providers.database.base import DatabaseDataProvider
from universal_transfer_operator.datasets.file.base import File
from universal_transfer_operator.datasets.table import Metadata, Table
from universal_transfer_operator.exceptions import DatabaseCustomError
from universal_transfer_operator.settings import (
    LOAD_TABLE_AUTODETECT_ROWS_COUNT,
    SNOWFLAKE_SCHEMA,
    SNOWFLAKE_STORAGE_INTEGRATION_AMAZON,
    SNOWFLAKE_STORAGE_INTEGRATION_GOOGLE,
)
from universal_transfer_operator.universal_transfer_operator import TransferIntegrationOptions

DEFAULT_STORAGE_INTEGRATION = {
    FileLocation.S3: SNOWFLAKE_STORAGE_INTEGRATION_AMAZON,
    FileLocation.GS: SNOWFLAKE_STORAGE_INTEGRATION_GOOGLE,
}

ASTRO_SDK_TO_SNOWFLAKE_FILE_FORMAT_MAP = {
    FileType.CSV: "CSV",
    FileType.NDJSON: "JSON",
    FileType.PARQUET: "PARQUET",
}

COPY_INTO_COMMAND_FAIL_STATUS = "LOAD_FAILED"

NATIVE_LOAD_SUPPORTED_FILE_TYPES = (FileType.CSV, FileType.NDJSON, FileType.PARQUET)
NATIVE_LOAD_SUPPORTED_FILE_LOCATIONS = (
    FileLocation.GS,
    FileLocation.S3,
)


@dataclass
class SnowflakeStage:
    """
    Dataclass which abstracts properties of a Snowflake Stage.

    Snowflake Stages are used to loading tables and unloading data from tables into files.

    Example:

    .. code-block:: python

        snowflake_stage = SnowflakeStage(
            name="stage_name",
            url="gcs://bucket/prefix",
            metadata=Metadata(database="SNOWFLAKE_DATABASE", schema="SNOWFLAKE_SCHEMA"),
        )

    .. seealso::
            `Snowflake official documentation on stage creation
            <https://docs.snowflake.com/en/sql-reference/sql/create-stage.html>`_
    """

    name: str = ""
    _name: str = field(init=False, repr=False, default="")
    url: str = ""
    metadata: Metadata = field(default_factory=Metadata)

    @staticmethod
    def _create_unique_name() -> str:
        """
        Generate a valid Snowflake stage name.

        :return: unique stage name
        """
        return (
            "stage_"
            + random.choice(string.ascii_lowercase)
            + "".join(random.choice(string.ascii_lowercase + string.digits) for _ in range(7))
        )

    def set_url_from_file(self, file: File) -> None:
        """
        Given a file to be loaded/unloaded to from Snowflake, identifies its folder and
        sets as self.url.

        It is also responsible for adjusting any path specific requirements for Snowflake.

        :param file: File to be loaded/unloaded to from Snowflake
        """
        # the stage URL needs to be the folder where the files are
        # https://docs.snowflake.com/en/sql-reference/sql/create-stage.html#external-stage-parameters-externalstageparams
        url = urlparse(file.location.snowflake_stage_path)
        url = url._replace(path=url.path[: url.path.rfind("/") + 1])
        url = url._replace(scheme="gcs") if url.scheme == "gs" else url
        self.url = url.geturl()

    @property  # type: ignore
    def name(self) -> str:
        """
        Return either the user-defined name or auto-generated one.

        :return: stage name
        :sphinx-autoapi-skip:
        """
        if not self._name:
            self._name = self._create_unique_name()
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        """
        Set the stage name.

        :param value: Stage name.
        """
        if not isinstance(value, property) and value != self._name:
            self._name = value

    @property
    def qualified_name(self) -> str:
        """
        Return stage qualified name. In Snowflake, it is the database, schema and table

        :return: Snowflake stage qualified name (e.g. database.schema.table)
        """
        qualified_name_lists = [
            self.metadata.database,
            self.metadata.schema,
            self.name,
        ]
        qualified_name = ".".join(name for name in qualified_name_lists if name)
        return qualified_name


@attr.define
class SnowflakeOptions(TransferIntegrationOptions):
    """
    Snowflake load option for naive transfer.
    """

    file_options: dict = attr.field(factory=dict)
    copy_options: dict = attr.field(factory=dict)
    storage_integration: str | None = attr.field(default=None)
    validation_mode: str | None = attr.field(default=None)


class SnowflakeDataProvider(DatabaseDataProvider):
    """SnowflakeDataProvider represent all the DataProviders interactions with Snowflake Databases."""

    DEFAULT_SCHEMA = SNOWFLAKE_SCHEMA
    OPTIONS_CLASS = SnowflakeOptions

    def __init__(
        self,
        dataset: Table,
        transfer_mode,
        transfer_params: SnowflakeOptions = SnowflakeOptions(),
    ):
        self.dataset = dataset
        self.transfer_params: SnowflakeOptions = transfer_params
        self.transfer_mode = transfer_mode
        self.transfer_mapping = set()
        self.LOAD_DATA_NATIVELY_FROM_SOURCE: dict = {}
        super().__init__(
            dataset=self.dataset, transfer_mode=self.transfer_mode, transfer_params=self.transfer_params
        )

    def __repr__(self):
        return f'{self.__class__.__name__}(conn_id="{self.dataset.conn_id})'

    @property
    def sql_type(self) -> str:
        return "snowflake"

    @property
    def hook(self) -> SnowflakeHook:
        """Retrieve Airflow hook to interface with the Snowflake database."""
        kwargs = {}
        _hook = SnowflakeHook(snowflake_conn_id=self.dataset.conn_id)
        if self.dataset and self.dataset.metadata:
            if _hook.database is None and self.dataset.metadata.database:
                kwargs.update({"database": self.dataset.metadata.database})
            if _hook.schema is None and self.dataset.metadata.schema:
                kwargs.update({"schema": self.dataset.metadata.schema})
        return SnowflakeHook(snowflake_conn_id=self.dataset.conn_id, **kwargs)

    @property
    def default_metadata(self) -> Metadata:
        """
        Fill in default metadata values for table objects addressing snowflake databases
        """
        connection = self.hook.get_conn()
        return Metadata(  # type: ignore
            schema=connection.schema,
            database=connection.database,
        )

    # ---------------------------------------------------------
    # Table metadata
    # ---------------------------------------------------------
    @staticmethod
    def get_table_qualified_name(table: Table) -> str:  # skipcq: PYL-R0201
        """
        Return table qualified name. In Snowflake, it is the database, schema and table

        :param table: The table we want to retrieve the qualified name for.
        """
        qualified_name_lists = [
            table.metadata.database,
            table.metadata.schema,
            table.name,
        ]
        qualified_name = ".".join(name for name in qualified_name_lists if name)
        return qualified_name

    def schema_exists(self, schema: str) -> bool:
        """
        Checks if a schema exists in the database

        :param schema: DB Schema - a namespace that contains named objects like (tables, functions, etc)
        """

        # Below code is added due to breaking change in apache-airflow-providers-snowflake==3.2.0,
        # we need to pass handler param to get the rows. But in version apache-airflow-providers-snowflake==3.1.0
        # if we pass the handler provider raises an exception AttributeError 'sfid'.
        try:
            schemas = self.hook.run(
                "SELECT SCHEMA_NAME from information_schema.schemata WHERE LOWER(SCHEMA_NAME) = %(schema_name)s;",
                parameters={"schema_name": schema.lower()},
                handler=lambda cur: cur.fetchall(),
            )
        except AttributeError:
            schemas = self.hook.run(
                "SELECT SCHEMA_NAME from information_schema.schemata WHERE LOWER(SCHEMA_NAME) = %(schema_name)s;",
                parameters={"schema_name": schema.lower()},
            )
        if schemas is None:
            raise ValueError("Query didn't returned list of schema")
        try:
            # Handle case for apache-airflow-providers-snowflake<4.0.1
            created_schemas = [x["SCHEMA_NAME"] for x in schemas]
        except TypeError:
            # Handle case for apache-airflow-providers-snowflake>=4.0.1
            created_schemas = [x[0] for x in schemas]
        return len(created_schemas) == 1

    def create_table_using_schema_autodetection(
        self,
        table: Table,
        file: File | None = None,
        dataframe: pd.DataFrame | None = None,
    ) -> None:  # skipcq PYL-W0613
        """
        Create a SQL table, automatically inferring the schema using the given file.
        Overriding default behaviour and not using the `prep_table` since it doesn't allow the adding quotes.

        :param table: The table to be created.
        :param file: File used to infer the new table columns.
        :param dataframe: Dataframe used to infer the new table columns if there is no file
        """
        if file is None:
            if dataframe is None:
                raise ValueError(
                    "File or Dataframe is required for creating table using schema autodetection"
                )
            source_dataframe = dataframe
        else:
            source_dataframe = file.export_to_dataframe(nrows=LOAD_TABLE_AUTODETECT_ROWS_COUNT)

        # We are changing the case of table name to ease out on the requirements to add quotes in raw queries.
        # ToDO - Currently, we cannot to append using load_file to a table name which is having name in lower case.
        pandas_tools.write_pandas(
            conn=self.hook.get_conn(),
            df=source_dataframe,
            table_name=table.name.upper(),
            schema=table.metadata.schema,
            database=table.metadata.database,
            chunk_size=DEFAULT_CHUNK_SIZE,
            quote_identifiers=self.use_quotes(source_dataframe),
            auto_create_table=True,
        )
        # We are truncating since we only expect table to be created with required schema.
        # Since this method is used by both native and pandas path we cannot skip this step.
        self.truncate_table(table)

    def load_pandas_dataframe_to_table(
        self,
        source_dataframe: pd.DataFrame,
        target_table: Table,
        if_exists: LoadExistStrategy = "replace",
        chunk_size: int = DEFAULT_CHUNK_SIZE,
    ) -> None:
        """
        Create a table with the dataframe's contents.
        If the table already exists, append or replace the content, depending on the value of `if_exists`.

        :param source_dataframe: Local or remote filepath
        :param target_table: Table in which the file will be loaded
        :param if_exists: Strategy to be used in case the target table already exists.
        :param chunk_size: Specify the number of rows in each batch to be written at a time.
        """
        self._assert_not_empty_df(source_dataframe)

        auto_create_table = False
        if not self.table_exists(target_table):
            auto_create_table = True
        elif if_exists == "replace":
            self.create_table(target_table, dataframe=source_dataframe)

        # We are changing the case of table name to ease out on the requirements to add quotes in raw queries.
        # ToDO - Currently, we cannot to append using load_file to a table name which is having name in lower case.
        pandas_tools.write_pandas(
            conn=self.hook.get_conn(),
            df=source_dataframe,
            table_name=target_table.name.upper(),
            schema=target_table.metadata.schema,
            database=target_table.metadata.database,
            chunk_size=chunk_size,
            quote_identifiers=self.use_quotes(source_dataframe),
            auto_create_table=auto_create_table,
        )

    def truncate_table(self, table):
        """Truncate table"""
        self.run_sql(f"TRUNCATE {self.get_table_qualified_name(table)}")

    @classmethod
    def use_quotes(cls, cols: Sequence[str]) -> bool:
        """
        With snowflake identifier we have two cases,

        1. When Upper/Mixed case col names are used
            We are required to preserver the text casing of the col names. By adding the quotes around identifier.
        2. When lower case col names are used
            We can use them as is

        This is done to be in sync with Snowflake SQLAlchemy dialect.
        https://docs.snowflake.com/en/user-guide/sqlalchemy.html#object-name-case-handling

        Snowflake stores all case-insensitive object names in uppercase text. In contrast, SQLAlchemy considers all
        lowercase object names to be case-insensitive. Snowflake SQLAlchemy converts the object name case during
        schema-level communication (i.e. during table and index reflection). If you use uppercase object names,
        SQLAlchemy assumes they are case-sensitive and encloses the names with quotes. This behavior will cause
        mismatches against data dictionary data received from Snowflake, so unless identifier names have been truly
        created as case sensitive using quotes (e.g. "TestDb"), all lowercase names should be used on the SQLAlchemy
        side.

        :param cols: list of columns
        """
        return any(col for col in cols if not col.islower() and not col.isupper())

    @property
    def openlineage_dataset_name(self) -> str:
        """
        Returns the open lineage dataset name as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        Example: db_name.schema_name.table_name
        """
        conn = self.hook.get_connection(self.dataset.conn_id)
        conn_extra = conn.extra_dejson
        schema = conn_extra.get("schema") or conn.schema
        db = conn_extra.get("database")
        return f"{db}.{schema}.{self.dataset.name}"

    @property
    def openlineage_dataset_namespace(self) -> str:
        """
        Returns the open lineage dataset namespace as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        Example: snowflake://ACCOUNT
        """
        account = self.hook.get_connection(self.dataset.conn_id).extra_dejson.get("account")
        return f"{self.sql_type}://{account}"

    @property
    def openlineage_dataset_uri(self) -> str:
        """
        Returns the open lineage dataset uri as per
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md
        """
        return f"{self.openlineage_dataset_namespace}{self.openlineage_dataset_name}"

    def is_native_path_available(
        self,
        source_dataset: File | Table,  # skipcq PYL-W0613, PYL-R0201
    ) -> bool:
        """
        Check if there is an optimised path for source to destination.

        :param source_dataset: File | Table from which we need to transfer data
        """
        if isinstance(source_dataset, Table):
            return False
        is_file_type_supported = source_dataset.type.name in NATIVE_LOAD_SUPPORTED_FILE_TYPES
        is_file_location_supported = (
            source_dataset.location.location_type in NATIVE_LOAD_SUPPORTED_FILE_LOCATIONS
        )
        return is_file_type_supported and is_file_location_supported

    def load_file_to_table_natively(
        self,
        source_file: File,
        target_table: Table,
        if_exists: LoadExistStrategy = "replace",
        native_support_kwargs: dict | None = None,
        **kwargs,
    ):  # skipcq PYL-W0613
        """
        Load the content of a file to an existing Snowflake table natively by:
        - Creating a Snowflake external stage
        - Using Snowflake COPY INTO statement

        Requirements:
        - The user must have permissions to create a STAGE in Snowflake.
        - If loading from GCP Cloud Storage, `native_support_kwargs` must define `storage_integration`
        - If loading from AWS S3, the credentials for creating the stage may be
        retrieved from the Airflow connection or from the `storage_integration`
        attribute within `native_support_kwargs`.

        :param source_file: File from which we need to transfer data
        :param target_table: Table to which the content of the file will be loaded to
        :param if_exists: Strategy used to load (currently supported: "append" or "replace")
        :param native_support_kwargs: may be used for the stage creation, as described above.

        .. seealso::
            `Snowflake official documentation on COPY INTO
            <https://docs.snowflake.com/en/sql-reference/sql/copy-into-table.html>`_
            `Snowflake official documentation on CREATE STAGE
            <https://docs.snowflake.com/en/sql-reference/sql/create-stage.html>`_

        """
        native_support_kwargs = native_support_kwargs or {}
        # if not self.load_options:
        #     self.load_options = SnowflakeLoadOptions()

        storage_integration = native_support_kwargs.get("storage_integration", None)
        if storage_integration is None:
            storage_integration = self.transfer_params.storage_integration

        stage = self.create_stage(file=source_file, storage_integration=storage_integration)

        rows = self._copy_into_table_from_stage(
            source_file=source_file, target_table=target_table, stage=stage
        )
        self.evaluate_results(rows)

    @staticmethod
    def _create_stage_auth_sub_statement(file: File, storage_integration: str | None = None) -> str:
        """
        Create authentication-related line for the Snowflake CREATE STAGE.
        Raise an exception if it is not defined.

        :param file: File to be copied from/to using stage
        :param storage_integration: Previously created Snowflake storage integration
        :return: String containing line to be used for authentication on the remote storage
        """
        storage_integration = storage_integration or DEFAULT_STORAGE_INTEGRATION.get(
            file.location.location_type
        )
        if storage_integration is not None:
            return f"storage_integration = {storage_integration};"
        return file.location.get_snowflake_stage_auth_sub_statement()  # type: ignore

    def create_stage(
        self,
        file: File,
        storage_integration: str | None = None,
        metadata: Metadata | None = None,
    ) -> SnowflakeStage:
        """
        Creates a new named external stage to use for loading data from files into Snowflake
        tables and unloading data from tables into files.

        At the moment, the following ways of authenticating to the backend are supported:
        * Google Cloud Storage (GCS): using storage_integration, previously created
        * Amazon (S3): one of the following:
        (i) using storage_integration or
        (ii) retrieving the AWS_KEY_ID and AWS_SECRET_KEY from the Airflow file connection

        :param file: File to be copied from/to using stage
        :param storage_integration: Previously created Snowflake storage integration
        :param metadata: Contains Snowflake database and schema information
        :return: Stage created

        .. seealso::
            `Snowflake official documentation on stage creation
            <https://docs.snowflake.com/en/sql-reference/sql/create-stage.html>`_
        """

        auth = self._create_stage_auth_sub_statement(file=file, storage_integration=storage_integration)

        # if not self.load_options:
        #     self.load_options = SnowflakeLoadOptions()
        metadata = metadata or self.default_metadata
        stage = SnowflakeStage(metadata=metadata)
        stage.set_url_from_file(file)

        file_format = ASTRO_SDK_TO_SNOWFLAKE_FILE_FORMAT_MAP[file.type.name]
        copy_options = []
        copy_options.extend([f"{k}={v}" for k, v in self.transfer_params.copy_options.items()])
        file_options = [f"{k}={v}" for k, v in self.transfer_params.file_options.items()]
        file_options.extend([f"TYPE={file_format}", "TRIM_SPACE=TRUE"])
        file_options_str = ", ".join(file_options)
        copy_options_str = ", ".join(copy_options)
        sql_statement = "".join(
            [
                f"CREATE OR REPLACE STAGE {stage.qualified_name} URL='{stage.url}' ",
                f"FILE_FORMAT=({file_options_str}) ",
                f"COPY_OPTIONS=({copy_options_str}) ",
                auth,
            ]
        )
        logging.debug("SQL statement executed: %s ", sql_statement)
        self.run_sql(sql_statement)

        return stage

    def _copy_into_table_from_stage(self, source_file, target_table, stage):
        table_name = self.get_table_qualified_name(target_table)
        file_path = os.path.basename(source_file.path) or ""
        sql_statement = f"COPY INTO {table_name} FROM @{stage.qualified_name}/{file_path}"

        self._validate_before_copy_into(source_file, target_table, stage)

        # Below code is added due to breaking change in apache-airflow-providers-snowflake==3.2.0,
        # we need to pass handler param to get the rows. But in version apache-airflow-providers-snowflake<=3.1.0
        # if we pass the handler provider raises an exception AttributeError
        try:
            if packaging_version.parse(
                version("apache-airflow-providers-snowflake")
            ) <= packaging_version.parse("3.1.0"):
                rows = self.hook.run(sql_statement)
            else:
                rows = self.hook.run(sql_statement, handler=lambda cur: cur.fetchall())
        except ValueError as exe:
            raise DatabaseCustomError from exe
        finally:
            self.drop_stage(stage)
        return rows

    def _validate_before_copy_into(
        self, source_file: File, target_table: Table, stage: SnowflakeStage
    ) -> None:
        """Validate COPY INTO command to tests the files for errors but does not load them."""
        if self.transfer_params:
            if self.transfer_params.validation_mode is None:
                return
            table_name = self.get_table_qualified_name(target_table)
            file_path = os.path.basename(source_file.path) or ""
            sql_statement = (
                f"COPY INTO {table_name} FROM "
                f"@{stage.qualified_name}/{file_path} VALIDATION_MODE='{self.transfer_params.validation_mode}'"
            )
            try:
                self.hook.run(sql_statement, handler=lambda cur: cur.fetchall())
            except ProgrammingError as load_exception:
                self.drop_stage(stage)
                raise load_exception

    def drop_stage(self, stage: SnowflakeStage) -> None:
        """
        Runs the snowflake query to drop stage if it exists.

        :param stage: Stage to be dropped
        """
        sql_statement = f"DROP STAGE IF EXISTS {stage.qualified_name};"
        self.hook.run(sql_statement, autocommit=True)

    @staticmethod
    def evaluate_results(rows):
        """check the error state returned by snowflake when running `copy into` query."""
        try:
            # Handle case for apache-airflow-providers-snowflake<4.0.1
            if any(row["status"] == COPY_INTO_COMMAND_FAIL_STATUS for row in rows):
                raise DatabaseCustomError(rows)
        except TypeError:
            # Handle case for apache-airflow-providers-snowflake>=4.0.1
            if any(row[0] == COPY_INTO_COMMAND_FAIL_STATUS for row in rows):
                raise DatabaseCustomError(rows)
